"""Data loading backend for katdal.

To avoid confusion, the term "baseline" is used for a pair of antennas, while
"correlation product" is used for a pair of single-pol inputs. katdal works
on correlation products while katsdpimager works on baselines, so some
conversion is needed.
"""

import argparse
import logging
import itertools
import bisect
import math
import urllib

import katdal
from katdal.lazy_indexer import DaskLazyIndexer
import numpy as np
from astropy import units
from astropy.time import Time
import astropy.io.fits

from . import polarization, loader_core, sky_model


_logger = logging.getLogger(__name__)


def _unique(seq):
    """Like np.unique, but operates in pure Python.

    This ensures that tuples and other objects preserve their identity instead
    of being coerced into numpy arrays.

    Parameters
    ----------
    seq : iterable
        Sequence of comparable objects

    Returns
    -------
    unique : list
        Sorted list of the unique objects from seq
    """
    data = sorted(seq)
    return [key for key, group in itertools.groupby(data)]


def _timestamp_to_fits(timestamp):
    """Convert :class:`katdal.Timestamp` or UNIX time to FITS header format."""
    return Time(float(timestamp), format='unix').utc.isot


class LoaderKatdal(loader_core.LoaderBase):
    def _find_target(self, target):
        """Find and return the target index based on the argument.

        Parameters
        ----------
        target : str or int or NoneType
            If None, autoselect the target. If an integer, treat it as a
            catalogue index. If a string, match it to a name in the
            catalogue.

        Raises
        ------
        ValueError
            If the selected target did not match the catalogue
        """
        if not self._file.catalogue:
            raise ValueError('The file does not contain any targets')
        if target is None:
            # Find first target with 'target' tag. If that fails, first without
            # bpcal or gaincal tag. If that fails too, first target.
            for i, trg in enumerate(self._file.catalogue):
                if 'target' in trg.tags:
                    return i
            for i, trg in enumerate(self._file.catalogue):
                if 'bpcal' not in trg.tags and 'gaincal' not in trg.tags:
                    return i
            return 0
        else:
            try:
                idx = int(target)
            except ValueError:
                for i, trg in enumerate(self._file.catalogue):
                    if target in [trg.name, trg.description] + trg.aliases:
                        return i
                raise ValueError('Target {} not found in catalogue'.format(target))
            else:
                if idx < 0 or idx >= len(self._file.catalogue):
                    raise ValueError('Target index {} is out of range'.format(idx))
                return idx

    def __init__(self, filename, options):
        super().__init__(filename, options)
        parser = argparse.ArgumentParser(
            prog='katdal options',
            usage='katdal options: [-i subarray=N] [-i spw=M] ...')
        parser.add_argument('--subarray', type=int, default=0,
                            help='Subarray index within file [%(default)s]')
        parser.add_argument('--spw', type=int, default=0,
                            help='Spectral window index within file [%(default)s]')
        parser.add_argument('--target', type=str,
                            help='Target to image (index or name) [auto]')
        parser.add_argument('--ref-ant', type=str, default='',
                            help='Override reference antenna for identifying scans [array]')
        parser.add_argument('--apply-cal', type=str, default='all',
                            help='Comma-separated calibration solutions to pre-apply [%(default)s]')
        parser.add_argument('--access-key', type=str, help='S3 access key')
        parser.add_argument('--secret-key', type=str, help='S3 secret key')
        args = parser.parse_args(options)

        open_args = dict(ref_ant=args.ref_ant, applycal=args.apply_cal)
        if (args.access_key is not None) != (args.secret_key is not None):
            raise ValueError('access-key and secret-key must be used together')
        if args.access_key is not None:
            open_args['credentials'] = (args.access_key, args.secret_key)

        self._file = katdal.open(filename, **open_args)
        if args.subarray < 0 or args.subarray >= len(self._file.subarrays):
            raise ValueError('Subarray {} is out of range'.format(args.subarray))
        if args.spw < 0 or args.spw >= len(self._file.spectral_windows):
            raise ValueError('Spectral window {} is out of range'.format(args.spw))
        self._spectral_window = self._file.spectral_windows[args.spw]
        target_idx = self._find_target(args.target)
        self._file.select(subarray=args.subarray, spw=args.spw,
                          targets=[target_idx], scans=['track'],
                          corrprods='cross')
        self._target = self._file.catalogue.targets[target_idx]
        _logger.info('Selected target %r', self._target.description)
        if self._target.body_type != 'radec':
            raise ValueError('Target does not have fixed RA/DEC')

        # Identify polarizations present in the file. Note that _unique
        # returns a sorted list.
        pols = _unique(a[-1] + b[-1] for a, b in self._file.corr_products)
        self._polarizations = pols

        # Compute permutation of correlation products to place all
        # polarizations for one baseline together, and identify missing
        # correlation product (so that they can be flagged). Note that if a
        # baseline is totally absent it is completely ignored, rather than
        # marked as missing.
        corr_product_inverse = {
            tuple(corr_product): i for i, corr_product in enumerate(self._file.corr_products)
        }
        # Find baselines represented by corr_products.
        baselines = _unique((a[:-1], b[:-1]) for a, b in self._file.corr_products)
        corr_product_permutation = []
        missing_corr_products = []
        for a, b in baselines:
            for pol in pols:
                corr_product = (a + pol[0], b + pol[1])
                corr_product_permutation.append(corr_product_inverse.get(corr_product, None))
        for i in range(len(corr_product_permutation)):
            if corr_product_permutation[i] is None:
                # Has to have some valid index for advanced indexing
                corr_product_permutation[i] = 0
                missing_corr_products.append(i)
        self._corr_product_permutation = corr_product_permutation
        self._missing_corr_products = missing_corr_products

        # Turn baselines from pairs of names into pairs of antenna indices
        ant_inverse = {antenna.name: i for i, antenna in enumerate(self._file.ants)}
        try:
            self._baselines = [(ant_inverse[a], ant_inverse[b]) for a, b in baselines]
        except KeyError:
            raise ValueError('File does not contain antenna specifications for all antennas')

        # Set up a reference antenna for the array centre. katdal's reference
        # antenna is the first antenna in the file, which is not as useful.
        # This determines the reference frame for UVW coordinates.
        self._ref_ant = self._file.sensor.get('Antennas/array/antenna')[0]
        corrections = ', '.join(self._file.applycal_products)
        if not corrections:
            corrections = 'none'
        _logger.info('Calibration corrections applied: %s', corrections)

    @classmethod
    def match(cls, filename):
        if filename.lower().endswith('.h5') or filename.lower().endswith('.rdb'):
            return True
        # katdal also supports URLs with query parameters, in which case the
        # URL string won't end with .rdb.
        try:
            url = urllib.parse.urlsplit(filename)
            return url.scheme == 'redis' or url.path.endswith('.rdb')
        except ValueError:
            # Invalid URL, but could still be valid for another loader
            return False

    def antenna_diameters(self):
        diameters = [ant.diameter for ant in self._file.ants]
        return units.Quantity(diameters, unit=units.m, dtype=np.float32)

    def antenna_positions(self):
        positions = [ant.position_ecef for ant in self._file.ants]
        return units.Quantity(positions, unit=units.m)

    def num_channels(self):
        return self._file.shape[1]

    def frequency(self, channel):
        return self._file.freqs[channel] * units.Hz

    def phase_centre(self):
        ra, dec = self._target.astrometric_radec()
        return units.Quantity([ra, dec], unit=units.rad)

    def polarizations(self):
        out_map = {
            'hh': polarization.STOKES_XX,
            'hv': polarization.STOKES_XY,
            'vh': polarization.STOKES_YX,
            'vv': polarization.STOKES_YY
        }
        return [out_map[pol] for pol in self._polarizations]

    def has_feed_angles(self):
        return True

    def data_iter(self, start_channel, stop_channel, max_chunk_vis=None):
        self._file.select(reset='F')
        n_file_times, n_file_chans, n_file_cp = self._file.shape
        self._file.select(channels=np.s_[start_channel : stop_channel])
        assert 0 <= start_channel < stop_channel <= n_file_chans
        n_chans = stop_channel - start_channel
        n_pols = len(self._polarizations)
        if max_chunk_vis is None:
            load_times = n_file_times
        else:
            load_times = max(1, max_chunk_vis // (n_chans * n_file_cp))
        # timestamps is a property, so ensure it's only evaluated once
        timestamps = self._file.timestamps
        baseline_idx = np.arange(len(self._baselines)).astype(np.int32)
        start = 0
        # Determine chunking scheme
        if isinstance(self._file.vis, DaskLazyIndexer):
            chunk_sizes = self._file.vis.dataset.chunks[0]
            chunk_boundaries = [0] + list(itertools.accumulate(chunk_sizes))
            if chunk_sizes and chunk_sizes[0] > load_times:
                _logger.warning('Chunk size is %d dumps but only %d loaded at a time. '
                                'Consider increasing --vis-load',
                                chunk_sizes[0], load_times)
        else:
            chunk_boundaries = list(range(n_file_times + 1))  # No chunk info available
        while start < n_file_times:
            end = min(n_file_times, start + load_times)
            # Align to chunking if possible
            aligned_end = chunk_boundaries[bisect.bisect(chunk_boundaries, end) - 1]
            if aligned_end > start:
                end = aligned_end
            # Load a chunk from the lazy indexer, then reindex to order
            # the baselines as desired.
            _logger.debug('Loading dumps %d:%d', start, end)
            select = np.s_[start:end, :, :]
            fix = np.s_[:, :, self._corr_product_permutation]
            if isinstance(self._file.vis, DaskLazyIndexer):
                vis, weights, flags = DaskLazyIndexer.get(
                    [self._file.vis, self._file.weights, self._file.flags], select)
            else:
                vis = self._file.vis[select]
                weights = self._file.weights[select]
                flags = self._file.flags[select]
            _logger.debug('Dumps %d:%d loaded', start, end)
            vis = vis[fix]
            weights = weights[fix]
            flags = flags[fix]
            # Flag missing correlation products
            if self._missing_corr_products:
                flags[:, :, self._missing_corr_products] = True
            # Apply flags to weights
            weights *= np.logical_not(flags)

            # Compute per-antenna UVW coordinates and parallactic angles.
            antenna_uvw = units.Quantity(self._target.uvw(
                self._file.ants, timestamp=timestamps[start:end], antenna=self._ref_ant))
            antenna_uvw = antenna_uvw.T   # Switch from (uvw, time, ant) to (ant, time, uvw)
            # parangle converts to degrees before returning, so we have to
            # convert back to radians.
            antenna_pa = units.Quantity(
                self._file.parangle[start:end, :].transpose(),
                unit=units.deg, dtype=np.float32, copy=False).to(units.rad)
            # We've mapped H to x and V to y, so we need the angle from x to H
            # rather than from x to V.
            antenna_pa -= math.pi / 2 * units.rad
            # Combine these into per-baseline UVW coordinates and feed angles
            uvw = np.empty((end - start, len(self._baselines), 3), np.float32)
            feed_angle1 = np.empty((end - start, len(self._baselines)), np.float32)
            feed_angle2 = np.empty_like(feed_angle1)
            for i, (a, b) in enumerate(self._baselines):
                uvw[:, i, :] = antenna_uvw[b] - antenna_uvw[a]
                feed_angle1[:, i] = antenna_pa[a]
                feed_angle2[:, i] = antenna_pa[b]

            # reshape everything into the target formats
            yield dict(
                uvw=uvw.reshape(-1, 3),
                weights=weights.swapaxes(0, 1).reshape(n_chans, -1, n_pols),
                baselines=np.tile(baseline_idx, end - start),
                vis=vis.swapaxes(0, 1).reshape(n_chans, -1, n_pols),
                feed_angle1=feed_angle1.reshape(-1),
                feed_angle2=feed_angle2.reshape(-1),
                progress=end,
                total=n_file_times)
            start = end

    def sky_model(self):
        try:
            source = self._file.source
        except AttributeError:
            raise sky_model.NoSkyModelError('This data set does not support sky models')

        return sky_model.KatpointSkyModel(sky_model.catalogue_from_telstate(
            source.telstate, source.capture_block_id, None, self._target))

    def extra_fits_headers(self):
        headers = astropy.io.fits.Header()
        timestamps = self._file.timestamps
        if not len(timestamps):
            avg = self._file.start_time.secs
        else:
            avg = np.mean(timestamps)

        headers['OBJECT'] = self._target.name
        headers['SPECSYS'] = 'TOPOCENT'
        # SSYSOBS is not needed because it defaults to TOPOCENT
        headers['DATE-OBS'] = _timestamp_to_fits(self._file.start_time)
        headers['DATE-AVG'] = _timestamp_to_fits(avg)
        headers['ONTIME'] = (len(timestamps) * self._file.dump_period,
                             '[s] Time tracking the target')
        if self._file.observer:
            headers['OBSERVER'] = self._file.observer
        if self._spectral_window.product:
            headers['INSTRUME'] = self._spectral_window.product

        try:
            array_ant = self._file.sensor['Antennas/array/antenna'][0]
            array_pos = array_ant.position_ecef
            headers['OBSGEO-X'] = array_pos[0]
            headers['OBSGEO-Y'] = array_pos[1]
            headers['OBSGEO-Z'] = array_pos[2]
        except (KeyError, IndexError):
            pass

        try:
            headers['HISTORY'] = f'Capture block id: {self._file.source.capture_block_id}'
        except AttributeError:
            pass
        try:
            headers['HISTORY'] = f'Stream name: {self._file.source.stream_name}'
        except AttributeError:
            pass

        return headers

    @property
    def raw_data(self):
        return self._file

    def close(self):
        # katdal does not provide a way to close an input file. The best we can
        # do is leave it to the garbage collector to sort out.
        self._file = None

    @property
    def raw_target(self):
        """Target as a katpoint object"""
        return self._target



"""Miscellaneous file format routines"""

import logging

import numpy as np
from astropy import units
import astropy.io.fits as fits
from astropy.time import Time

import katsdpimager
import katsdpimager.polarization as polarization


_logger = logging.getLogger(__name__)


#: Maps internal polarization constants to those used in FITS
# FITS swaps the meaning of X and Y relative to IEEE definitions (see AIPS
# memo 114).
_FITS_POLARIZATIONS = {
    polarization.STOKES_I: 1,
    polarization.STOKES_Q: 2,
    polarization.STOKES_U: 3,
    polarization.STOKES_V: 4,
    polarization.STOKES_RR: -1,
    polarization.STOKES_LL: -2,
    polarization.STOKES_RL: -3,
    polarization.STOKES_LR: -4,
    polarization.STOKES_YY: -5,
    polarization.STOKES_XX: -6,
    polarization.STOKES_YX: -7,
    polarization.STOKES_XY: -8
}


def _fits_polarizations(header, axis, polarizations):
    """Add keywords to a FITS header describing the polarization axis.

    Parameters
    ----------
    header : :py:class:`astropy.io.fits.Header`
        Header which will be updated
    axis : int
        Axis number containing the polarizations, in FITS enumeration
    polarizations : sequence
        List of polarization enums

    Returns
    -------
    ndarray
        Permutation that must be applied to the axis corresponding to
        `polarizations` to make things match.

    Raises
    ------
    ValueError
        If the set of `polarizations` cannot be represented as a linear
        transform in the FITS header.
    """

    # FITS and CASA enumerate polarizations differently. It might be possible
    # to provide an arbitrary mapping from pixel coordinates to world
    # coordinates, but the common cases can all be handled by a permutation
    # and a linear transformation.
    polarizations = np.array([_FITS_POLARIZATIONS[i] for i in polarizations])
    if polarizations[0] >= 0:
        pol_permute = np.argsort(polarizations)
    else:
        # FITS numbers non-IQUV polarizations in a decreasing manner
        pol_permute = np.argsort(-polarizations)
    polarizations = polarizations[pol_permute]
    pol_ref = polarizations[0]
    if len(polarizations) > 1:
        pol_delta = polarizations[1] - polarizations[0]
    else:
        pol_delta = 1
    if np.any(polarizations != np.arange(len(polarizations)) * pol_delta + pol_ref):
        raise ValueError('Polarizations do not form a linear sequence in FITS enumeration')
    header['CTYPE{}'.format(axis)] = 'STOKES'
    header['CRPIX{}'.format(axis)] = 1.0
    header['CRVAL{}'.format(axis)] = float(pol_ref)
    header['CDELT{}'.format(axis)] = float(pol_delta)
    return pol_permute


def write_fits_image(dataset, image, image_parameters, filename, channel,
                     beam=None, bunit='Jy/beam'):
    """Write an image to a FITS file.

    Parameters
    ----------
    dataset : :class:`katsdpimager.loader_core.LoaderBase`
        Source dataset (used to set metadata such as phase centre)
    image : :class:`numpy.ndarray`
        Image data in Jy/beam, indexed by polarization, m, l. For
        a 2M x 2N image, the phase centre is at coordinates (M, N).
    image_parameters : :class:`katsdpimager.parameters.ImageParameters`
        Metadata associated with the image
    filename : str
        File to write. It is silently overwritten if already present.
    channel : int
        Channel number to substitute into `filename` with printf formatting.
    beam : :class:`katsdpimager.beam.Beam`, optional
        Synthesized beam model to write to the header
    bunit : str, optional
        Value for the ``BUNIT`` header in the file. It can be explicitly set
        to ``None`` to avoid writing this key.

    Raises
    ------
    ValueError
        If the set of `polarizations` cannot be represented as a linear
        transform in the FITS header.
    """
    header = fits.Header()
    if bunit is not None:
        header['BUNIT'] = bunit
    header['ORIGIN'] = 'katsdpimager'
    header['HISTORY'] = f'Created by katsdpimager {katsdpimager.__version__}'
    header['TIMESYS'] = 'UTC'
    header['DATE'] = Time.now().utc.isot

    # Transformation from pixel coordinates to intermediate world coordinates,
    # which are taken to be l, m coordinates. The reference point is currently
    # taken to be the centre of the image (actually half a pixel beyond the
    # centre, because of the way fftshift works).  Note that astropy.io.fits
    # reverses the axis order. The X coordinate is computed differently
    # because the X axis is flipped to allow RA to increase right-to-left.
    header['CRPIX1'] = image.shape[2] * 0.5
    header['CRPIX2'] = image.shape[1] * 0.5 + 1.0
    header['CRPIX4'] = 1.0
    # FITS uses degrees; and RA increases right-to-left
    delt = np.arcsin(image_parameters.pixel_size).to(units.deg).value
    header['CDELT1'] = -delt
    header['CDELT2'] = delt
    header['CDELT4'] = 1.0

    # Transformation from intermediate world coordinates to world
    # coordinates (celestial coordinates in this case).
    # TODO: get equinox from input
    phase_centre = dataset.phase_centre()
    header['EQUINOX'] = 2000.0
    header['RADESYS'] = 'FK5'   # Julian equinox
    header['CUNIT1'] = 'deg'
    header['CUNIT2'] = 'deg'
    header['CUNIT4'] = 'Hz'
    header['CTYPE1'] = 'RA---SIN'
    header['CTYPE2'] = 'DEC--SIN'
    header['CTYPE4'] = 'FREQ    '
    header['CRVAL1'] = phase_centre[0].to(units.deg).value
    header['CRVAL2'] = phase_centre[1].to(units.deg).value
    header['CRVAL4'] = image_parameters.wavelength.to(
        units.Hz, equivalencies=units.spectral()).value
    if beam is not None:
        major = beam.major * image_parameters.pixel_size * units.rad
        minor = beam.minor * image_parameters.pixel_size * units.rad
        header['BMAJ'] = major.to(units.deg).value
        header['BMIN'] = minor.to(units.deg).value
        header['BPA'] = beam.theta.to(units.deg).value
    _fits_polarizations(header, 3, image_parameters.polarizations)
    header['DATAMIN'] = float(np.nanmin(image))
    header['DATAMAX'] = float(np.nanmax(image))

    header.update(dataset.extra_fits_headers())

    # l axis is reversed, because RA increases right-to-left.
    # Explicitly converting to big-endian has two advantages:
    # 1. It returns a contiguous array, which allows for a much faster path
    #    through writeto.
    # 2. If the image is little-endian, then astropy.io.fits will convert to
    #    big endian, write, and convert back again.
    # The disadvantage is an increase in memory usage.
    #
    # The np.newaxis adds an axis for frequency. While it's not required for
    # the FITS file to be valid (it's legal for the WCS transformations to
    # reference additional virtual axes), aplpy 1.1.1 doesn't handle it
    # (https://github.com/aplpy/aplpy/issues/350).
    image = np.require(image[np.newaxis, :, :, ::-1], image.dtype.newbyteorder('>'), 'C')
    hdu = fits.PrimaryHDU(image, header)
    hdu.writeto(filename, overwrite=True)


def _split_array(x, dtype):
    """Return a view of x which has one extra dimension. Each element in x is
    treated as some number of elements of type `dtype`, whose size must divide
    into the element size of `x`."""
    in_dtype = x.dtype
    out_dtype = np.dtype(dtype)
    if in_dtype.hasobject or out_dtype.hasobject:
        raise ValueError('dtypes containing objects are not supported')
    if in_dtype.itemsize % out_dtype.itemsize != 0:
        raise ValueError('item size does not evenly divide')

    interface = dict(x.__array_interface__)
    if interface.get('mask', None) is not None:
        raise ValueError('masked arrays are not supported')
    interface['shape'] = x.shape + (in_dtype.itemsize // out_dtype.itemsize,)
    if interface['strides'] is not None:
        interface['strides'] = x.strides + (out_dtype.itemsize,)
    interface['typestr'] = out_dtype.str
    interface['descr'] = out_dtype.descr
    return np.asarray(np.lib.stride_tricks.DummyArray(interface, base=x))


def write_fits_grid(grid, image_parameters, filename, channel):
    """Writes a UV grid to a FITS file.

    Parameters
    ----------
    grid : ndarray of complex
        Grid data indexed by polarization, m, l
    image_parameters : :class:`katsdpimager.parameters.ImageParameters`
        Metadata used to set headers
    filename : str
        File to write. It is silently overwritten if already present.
    channel : int
        Channel number to substitute into `filename` with printf formatting.

    Raises
    ------
    ValueError
        If the set of `polarizations` cannot be represented as a linear
        transform in the FITS header.
    """
    grid = _split_array(grid, image_parameters.real_dtype)
    grid = grid.transpose(3, 0, 1, 2)

    header = fits.Header()
    header['BUNIT'] = 'JY'
    header['ORIGIN'] = 'katsdpimager'
    header['CUNIT1'] = 'm'
    header['CRPIX1'] = grid.shape[3] // 2 + 1.0
    header['CRVAL1'] = 0.0
    header['CDELT1'] = float(image_parameters.cell_size / units.m)
    header['CUNIT2'] = 'm'
    header['CRPIX2'] = grid.shape[2] // 2 + 1.0
    header['CRVAL2'] = 0.0
    header['CDELT2'] = float(image_parameters.cell_size / units.m)
    pol_permute = _fits_polarizations(header, 3, image_parameters.polarizations)
    header['CTYPE4'] = 'COMPLEX'
    header['CRPIX4'] = 1.0
    header['CRVAL4'] = 1.0
    header['CDELT4'] = 1.0

    hdu = fits.PrimaryHDU(grid[:, pol_permute, :, :], header)
    hdu.writeto(filename, overwrite=True)



"""Data loading frontend"""

import warnings


_loader_classes = []


def _register_loader(loader_class):
    _loader_classes.append(loader_class)


def load(filename, options):
    for cls in _loader_classes:
        if cls.match(filename):
            return cls(filename, options)
    raise ValueError('No loader class is registered for "{}"'.format(filename))


def data_iter(dataset, vis_limit, vis_load, start_channel, stop_channel):
    """Wrapper around :py:meth:`katsdpimager.loader_core.LoaderBase.data_iter`
    that handles truncation to a number of visibilities specified on the
    command line.
    """
    N = vis_limit
    it = dataset.data_iter(start_channel, stop_channel, vis_load)
    for chunk in it:
        if N is not None:
            if N < len(chunk['uvw']):
                for key in ['uvw', 'baselines']:
                    if key in chunk:
                        chunk[key] = chunk[key][:N]
                for key in ['weights', 'vis']:
                    if key in chunk:
                        chunk[key] = chunk[key][:, :N]
                chunk['progress'] = chunk['total']
        yield chunk
        if N is not None:
            N -= len(chunk['uvw'])
            if N == 0:
                it.close()
                return


try:
    import katsdpimager.loader_ms
    _register_loader(katsdpimager.loader_ms.LoaderMS)
except ImportError:
    warnings.warn("Failed to load loader_ms. Possibly python-casacore is missing or broken.",
                  ImportWarning)

try:
    import katsdpimager.loader_katdal
    _register_loader(katsdpimager.loader_katdal.LoaderKatdal)
except ImportError:
    warnings.warn("Failed to load loader_katdal. Possibly katdal is missing or broken.",
                  ImportWarning)



"""Utilities for test code"""

import numpy as np


class RandomState(np.random.RandomState):
    """Extension of :class:`random.RandomState` with extra distributions"""
    def complex_normal(self, loc=0.0j, scale=1.0, size=None):
        """Circularly symmetric Gaussian in the Argand plane"""
        return self.normal(np.real(loc), scale, size) + 1j * self.normal(np.imag(loc), scale, size)

    def complex_uniform(self, low=0.0, high=1.0, size=None):
        """Uniform distribution over a rectangle in the complex plane.

        If low or high is purely real (by dtype, not by value), it is taken to
        be the boundary on both the real and imaginary extent.
        """
        if not np.iscomplexobj(low):
            low = np.asarray(low) * (1 + 1j)
        if not np.iscomplexobj(high):
            high = np.asarray(high) * (1 + 1j)
        return self.uniform(np.real(low), np.real(high), size) \
            + 1j * self.uniform(np.imag(low), np.imag(high), size)



#!/usr/bin/env python

"""Benchmark suite for various steps in imaging"""

import argparse
import timeit
import json

import katpoint
import pkg_resources
import numpy as np
from astropy import units
from katsdpsigproc import accel

from katsdpimager import grid, preprocess, parameters, polarization, fft


def load_antennas():
    """Return an array of katpoint antenna objects"""
    antennas = []
    with pkg_resources.resource_stream(__name__, 'meerkat_antennas.txt') as f:
        for line in f:
            antennas.append(katpoint.Antenna(line))
    return antennas


def add_parameters(args):
    """Augment args with extra fields for parameter objects"""
    grid_oversample = 8
    antialias_width = 7
    # This is used to pick the number of w slices. That allows the --kernel-width
    # argument to vary ONLY the kernel width used in gridding, without
    # simultaneously affecting the set of visibilities used.
    nominal_kernel_width = 60
    if hasattr(args, 'kernel_width'):
        kernel_width = args.kernel_width
    else:
        kernel_width = nominal_kernel_width
    eps_w = 0.001
    antennas = load_antennas()
    longest = 0.0
    for i in range(len(antennas)):
        for j in range(i):
            longest = max(longest, np.linalg.norm(antennas[i].baseline_toward(antennas[j])))
    longest = longest * units.m
    array_parameters = parameters.ArrayParameters(antennas[0].diameter * units.m, longest)
    image_parameters = parameters.ImageParameters(
        q_fov=1.0,
        image_oversample=5,
        frequency=args.frequency * units.Hz,
        array=array_parameters,
        polarizations=polarization.STOKES_IQUV[:args.polarizations],
        dtype=np.float32)

    grid_oversample = 8
    w_step = image_parameters.cell_size / grid_oversample
    w_slices = parameters.w_slices(
        image_parameters,
        longest,
        eps_w,
        nominal_kernel_width,
        antialias_width)
    w_planes = int(np.ceil(longest / w_step / w_slices))
    grid_parameters = parameters.GridParameters(
        antialias_width=antialias_width,
        oversample=grid_oversample,
        image_oversample=4,
        w_slices=w_slices,
        w_planes=w_planes,
        max_w=longest,
        kernel_width=kernel_width)
    args.antennas = antennas
    args.array_parameters = array_parameters
    args.grid_parameters = grid_parameters
    args.image_parameters = image_parameters


def make_uvw(args, n_time):
    start = 946728000.0  # J2000, in UNIX time
    dec = -np.pi / 4
    target = katpoint.construct_radec_target(0, dec)
    timestamps = np.arange(n_time) * args.int_time + start
    ref = katpoint.Antenna('', *args.antennas[0].ref_position_wgs84)
    basis = target.uvw_basis(timestamp=timestamps, antenna=ref)
    antenna_uvw = []
    rows = [[], [], []]
    for antenna in args.antennas:
        enu = np.array(ref.baseline_toward(antenna))
        antenna_uvw.append(np.tensordot(basis, enu, ([1], [0])))
    for i in range(len(args.antennas)):
        for j in range(i):
            u, v, w = antenna_uvw[j] - antenna_uvw[i]
            rows[0].append(u)
            rows[1].append(v)
            rows[2].append(w)
    uvw = np.array(rows, dtype=np.float32)
    return uvw.reshape(3, uvw.shape[1] * n_time).T.copy()


def make_vis(args, n_time):
    """Generate uncompressed visibilities

    Parameters
    ----------
    args : argparse.Namespace
        Command-line arguments
    n_time : int
        Number of time samples
    """
    n_baselines = len(args.antennas) * (len(args.antennas) - 1) // 2
    N = n_time * n_baselines
    uvw = make_uvw(args, n_time)
    weights = np.ones((1, N, args.polarizations), np.float32)
    vis = np.ones((1, N, args.polarizations), np.complex64)
    baselines = np.repeat(np.arange(n_baselines, dtype=np.int32), n_time)
    return uvw, weights, baselines, vis


def make_compressed_vis(args, n_time):
    """Generate compressed visibilities

    Parameters
    ----------
    args : argparse.Namespace
        Command-line arguments
    n_time : int
        Number of time samples

    Returns
    -------
    reader : :py:class:`katsdpimager.preprocess.VisibilityReader`
        Reader that iterates the visibilities
    """
    uvw, weights, baselines, vis = make_vis(args, n_time)
    if args.write:
        collector = preprocess.VisibilityCollectorHDF5(
            args.write,
            [args.image_parameters], [args.grid_parameters], vis.shape[1])
    else:
        collector = preprocess.VisibilityCollectorMem(
            [args.image_parameters], [args.grid_parameters], vis.shape[1])
    mueller = np.identity(args.polarizations, np.complex64)
    collector.add(uvw, weights, baselines, vis, None, None, mueller, None)
    collector.close()
    reader = collector.reader()
    return reader


def benchmark_preprocess(args):
    add_parameters(args)
    n_time = 900
    uvw, weights, baselines, vis = make_vis(args, n_time)
    n_vis = vis.shape[0] * vis.shape[1]
    start = timeit.default_timer()
    collector = preprocess.VisibilityCollectorMem(
        [args.image_parameters], [args.grid_parameters], n_vis)
    mueller = np.identity(args.polarizations, np.complex64)
    collector.add(uvw, weights, baselines, vis, None, None, mueller, None)
    collector.close()
    end = timeit.default_timer()
    elapsed = end - start
    print("preprocessed {} visibilities in {} seconds".format(n_vis, elapsed))
    print("{:.2f} Mvis/s".format(n_vis / elapsed / 1e6))


def benchmark_grid_degrid(args):
    n_time = 3600
    add_parameters(args)
    N = n_time * len(args.antennas) * (len(args.antennas) - 1) // 2
    reader = make_compressed_vis(args, n_time)

    context = accel.create_some_context()
    queue = context.create_tuning_command_queue()
    gridder_template = args.template_class(context, args.image_parameters,
                                           args.grid_parameters, tuning=args.tuning)
    gridder = gridder_template.instantiate(queue, args.array_parameters, N)
    gridder.ensure_all_bound()
    elapsed = 0.0
    N_compressed = 0
    uv = gridder.buffer('uv').empty_like()
    w_plane = gridder.buffer('w_plane').empty_like()
    vis = gridder.buffer('vis').empty_like()
    for w_slice in range(reader.num_w_slices(0)):
        gridder.num_vis = reader.len(0, w_slice)
        N_compressed += gridder.num_vis
        if gridder.num_vis > 0:
            gridder.buffer('grid').zero(queue)
            start = 0
            for chunk in reader.iter_slice(0, w_slice):
                rng = slice(start, start + len(chunk))
                uv[rng, 0:2] = chunk['uv']
                uv[rng, 2:4] = chunk['sub_uv']
                w_plane[rng] = chunk['w_plane']
                vis[rng] = chunk['vis']
                start += len(chunk)
            gridder.buffer('uv').set_async(queue, uv)
            gridder.buffer('w_plane').set_async(queue, w_plane)
            gridder.buffer('vis').set_async(queue, vis)
            queue.finish()
            queue.start_tuning()
            gridder()
            elapsed += queue.stop_tuning()
            queue.finish()
    gaps = N_compressed * args.grid_parameters.kernel_width**2 * args.polarizations / elapsed
    print('Processed {} ({}) visibilities in {:.6f}s with kernel size {} and {} polarizations'
          .format(N_compressed, N, elapsed, args.grid_parameters.kernel_width, args.polarizations))
    print('{:.3f} GGAPS uncompressed'.format(gaps * N / N_compressed / 1e9))
    print('{:.3f} GGAPS compressed'.format(gaps / 1e9))


def benchmark_fft(args):
    context = accel.create_some_context()
    queue = context.create_tuning_command_queue()
    allocator = accel.SVMAllocator(context)
    shape = (args.pixels, args.pixels)
    template = fft.FftTemplate(queue, 2, shape, np.complex64, np.complex64, shape, shape)
    fn = template.instantiate(args.mode, allocator=allocator)
    fn.ensure_all_bound()
    # Zero-fill, just to ensure no NaNs etc
    fn.buffer('src').fill(0)
    fn.buffer('dest').fill(0)
    fn()  # Warm-up and forces data transfer
    queue.start_tuning()
    fn()
    elapsed = queue.stop_tuning()
    print('{pixels}x{pixels} in {elapsed:.6f} seconds'.format(pixels=args.pixels, elapsed=elapsed))
    # 8 bytes for complex64, 4 accesses (from source, to/from scratch, to dest)
    mem_rate = args.pixels * args.pixels * 8 * 4 / elapsed
    print('{:.3f} GiB/s'.format(mem_rate / 1024**3))


def add_arguments(subparser, arguments):
    arg_map = {
        '--polarizations': lambda: subparser.add_argument(
            '--polarizations', type=int, default=4, choices=[1, 2, 3, 4],
            help='Number of polarizations'),
        '--frequency': lambda: subparser.add_argument(
            '--frequency', type=float, default=1412000000.0,
            help='Observation frequency (Hz)'),
        '--int-time': lambda: subparser.add_argument(
            '--int-time', type=float, default=2.0, help='Integration time (seconds)'),
        '--kernel-width': lambda: subparser.add_argument(
            '--kernel-width', type=int, default=60, help='Convolutional kernel size in pixels'),
        '--pixels': lambda: subparser.add_argument(
            '--pixels', type=int, default=4608, help='Grid/image dimensions'),
        '--tuning': lambda: subparser.add_argument(
            '--tuning', type=json.loads, help='Tuning arguments (JSON)'),
        '--write': lambda: subparser.add_argument(
            '--write', type=str, help='Write compressed visibilities to HDF5 file')
    }
    for arg_name in arguments:
        arg_map[arg_name]()


def main():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(help='sub-commands')

    parser_preprocess = subparsers.add_parser('preprocess', help='preprocessing benchmark')
    parser_preprocess.set_defaults(func=benchmark_preprocess)
    add_arguments(parser_preprocess, ['--polarizations', '--frequency', '--int-time'])

    grid_degrid_arguments = ['--polarizations', '--frequency', '--int-time', '--kernel-width',
                             '--tuning', '--write']
    parser_grid = subparsers.add_parser('grid', help='gridding benchmark')
    parser_grid.set_defaults(func=benchmark_grid_degrid, template_class=grid.GridderTemplate)
    add_arguments(parser_grid, grid_degrid_arguments)

    parser_degrid = subparsers.add_parser('degrid', help='degridding benchmark')
    parser_degrid.set_defaults(func=benchmark_grid_degrid, template_class=grid.DegridderTemplate)
    add_arguments(parser_degrid, grid_degrid_arguments)

    parser_ifft = subparsers.add_parser('ifft', help='grid-to-image FFT benchmark')
    add_arguments(parser_ifft, ['--pixels'])
    parser_ifft.set_defaults(func=benchmark_fft, mode=fft.FFT_INVERSE)

    parser_fft = subparsers.add_parser('fft', help='image-to-grid FFT benchmark')
    add_arguments(parser_fft, ['--pixels'])
    parser_fft.set_defaults(func=benchmark_fft, mode=fft.FFT_FORWARD)

    args = parser.parse_args()
    args.func(args)


if __name__ == '__main__':
    main()



"""Utilities for converting type names between numpy and C"""

import numpy as np


def dtype_to_ctype(dtype):
    """Convert a numpy dtype to a C type suitable for CUDA or OpenCL.

    Only float32/64, complex64/128 and int32/uint32 are supported for now.
    """
    if dtype == np.float32:
        return 'float'
    elif dtype == np.complex64:
        return 'float2'
    elif dtype == np.float64:
        return 'double'
    elif dtype == np.complex128:
        return 'double2'
    elif dtype == np.int32:
        return 'int'
    elif dtype == np.uint32:
        return 'uint'
    else:
        raise ValueError('Unrecognised dtype {}'.format(dtype))


def real_to_complex(dtype):
    """Convert a real type to its complex equivalent"""
    if dtype == np.float32:
        return np.complex64
    elif dtype == np.float64:
        return np.complex128
    else:
        raise ValueError('Unrecognised dtype {}'.format(dtype))


def complex_to_real(dtype):
    """Convert a complex type to its real equivalent"""
    if dtype == np.complex64:
        return np.float32
    elif dtype == np.complex128:
        return np.float64
    else:
        raise ValueError('Unrecognised dtype {}'.format(dtype))



#!/usr/bin/env python3
import sys
import argparse
import logging
from contextlib import closing, contextmanager

import numpy as np
import colors
import katsdpsigproc.accel as accel

from katsdpimager import frontend, loader, io, progress, numba


logger = logging.getLogger()


class Writer(frontend.Writer):
    def __init__(self, args):
        self.args = args

    def write_fits_image(self, name, description, dataset, image, image_parameters, channel,
                         beam=None, bunit='JY/BEAM'):
        if name == 'clean':
            filename = self.args.output_file
        else:
            filename = getattr(self.args, 'write_' + name)
        if filename is not None:
            if '%' in filename:
                filename = filename % channel
            with progress.step('Write {}'.format(description)):
                io.write_fits_image(dataset, image, image_parameters, filename,
                                    channel, beam, bunit)

    def write_fits_grid(self, name, description, fftshift, grid_data, image_parameters, channel):
        filename = getattr(self.args, 'write_' + name)
        if filename is not None:
            if '%' in filename:
                filename = filename % channel
            with progress.step('Write {}'.format(description)):
                if fftshift:
                    grid_data = np.fft.fftshift(grid_data, axes=(1, 2))
                io.write_fits_grid(grid_data, image_parameters, filename, channel)


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('input_file', type=str, metavar='INPUT',
                        help='Input data set')
    parser.add_argument('output_file', type=str, metavar='OUTPUT',
                        help='Output FITS file')
    parser.add_argument('--log-level', type=str, default='INFO', metavar='LEVEL',
                        help='Logging level [%(default)s]')

    frontend.add_options(parser)

    group = parser.add_argument_group('Debugging options')
    group.add_argument('--host', action='store_true',
                       help='Perform operations on the CPU')
    group.add_argument('--write-weights', metavar='FILE',
                       help='Write imaging weights to FITS file')
    group.add_argument('--write-psf', metavar='FILE',
                       help='Write image of PSF to FITS file')
    group.add_argument('--write-grid', metavar='FILE',
                       help='Write UV grid to FITS file')
    group.add_argument('--write-dirty', metavar='FILE',
                       help='Write dirty image to FITS file')
    group.add_argument('--write-model', metavar='FILE',
                       help='Write model image to FITS file')
    group.add_argument('--write-residuals', metavar='FILE',
                       help='Write image residuals to FITS file')
    group.add_argument('--vis-limit', type=int, metavar='N',
                       help='Use only the first N visibilities')
    return parser


class DummyCommandQueue:
    """Stub equivalent to CUDA/OpenCL CommandQueue objects, that just provides
    an empty :meth:`finish` method."""
    def finish(self):
        pass


@contextmanager
def dummy_context():
    """Do-nothing context manager used in place of a device context."""
    yield


class ColorFormatter(logging.Formatter):
    COLORS = {
        logging.CRITICAL: colors.red,
        logging.ERROR: colors.red,
        logging.WARNING: colors.magenta,
        logging.INFO: colors.green,
        logging.DEBUG: colors.cyan
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def format(self, record):
        msg = super().format(record)
        if record.levelno in self.COLORS:
            msg = self.COLORS[record.levelno](msg)
        return msg


def configure_logging(args):
    log_handler = logging.StreamHandler()
    fmt = "[%(levelname)s] %(message)s"
    if sys.stderr.isatty():
        log_handler.setFormatter(ColorFormatter(fmt))
    else:
        log_handler.setFormatter(logging.Formatter(fmt))
    # In case some other module has decided to configure logging, remove any
    # existing handlers.
    for handler in list(logger.handlers):
        logger.removeHandler(handler)
    logger.addHandler(log_handler)
    logger.setLevel(args.log_level.upper())


def main():
    parser = get_parser()
    args = parser.parse_args()
    args.input_option = ['--' + opt for opt in args.input_option]
    if (args.stop_channel is None or args.stop_channel - args.start_channel > 1):
        if '%' not in args.output_file:
            parser.error('More than one channel selected but no %d in output filename')
    configure_logging(args)

    queue = None
    context = None
    if not args.host:
        context = accel.create_some_context(device_filter=lambda x: x.is_cuda)
        queue = context.create_command_queue()
    else:
        context = dummy_context()
        queue = DummyCommandQueue()
        if not numba.have_numba:
            logger.warning('could not import numba: --host mode will be VERY slow')

    with closing(loader.load(args.input_file, args.input_option)) as dataset:
        frontend.run(args, context, queue, dataset, Writer(args))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""Base classes used by loader modules"""
import numpy as np
from astropy import units
import astropy.io.fits

from . import parameters, sky_model


class LoaderBase:
    def __init__(self, filename, options):
        """Open the file"""
        self.filename = filename

    @classmethod
    def match(cls, filename):
        """Return True if this loader can handle the file"""
        raise NotImplementedError('Abstract base class')

    def antenna_diameters(self):
        """Effective diameters of the antennas."""
        raise NotImplementedError('Abstract base class')

    def antenna_diameter(self):
        """Return the common diameter of all dishes.

        Raises
        ------
        ValueError
            if the dishes do not all have the same diameter
        """
        diameters = self.antenna_diameters()
        D = diameters[0]
        if not np.all(diameters == D):
            raise ValueError('Diameters are not all equal')
        return D

    def antenna_positions(self):
        """Return the antenna positions. The coordinate system is arbitrary
        since it is used only to compute baseline lengths.

        Returns
        -------
        Quantity, shape (n, 3)
            XYZ coordinates for each antenna.
        """
        raise NotImplementedError('Abstract base class')

    def longest_baseline(self):
        """Return the length of the longest baseline."""
        positions = self.antenna_positions()
        ans = 0.0 * units.m
        for i in range(len(positions)):
            for j in range(i):
                baseline = positions[i] - positions[j]
                ans = max(ans, np.sqrt(baseline.dot(baseline)))
        return ans

    def array_parameters(self):
        """Return ArrayParameters object for the array used in the observation."""
        return parameters.ArrayParameters(
            self.antenna_diameter(), self.longest_baseline())

    def num_channels(self):
        """Return total number of channels, which are assumed to be contiguous."""
        raise NotImplementedError('Abstract base class')

    def frequency(self, channel):
        """Return frequency for a given channel.

        Returns
        -------
        Quantity
            Frequency, in appropriate reference frame for transforming UVW
            coordinates from distance to wavelength count.
        """
        raise NotImplementedError('Abstract base class')

    def phase_centre(self):
        """Return direction corresponding to l=0, m=0.

        Returns
        -------
        Quantity of shape (2,)
            RA and DEC for phase centre, in J2000 epoch.

            TODO: use katpoint or astropy.coordinates quantity instead?
        """
        raise NotImplementedError('Abstract base class')

    def polarizations(self):
        """Return polarizations stored in the data.

        Returns
        -------
        list
            List of polarization constants from
            :py:mod:`katsdpsigproc.parameters`.
        """
        raise NotImplementedError('Abstract base class')

    def has_feed_angles(self):
        """Return whether the data iterator will return `feed_angle1` and
        `feed_angle2`.
        """
        raise NotImplementedError('Abstract base class')

    def data_iter(self, start_channel, stop_channel, max_chunk_vis=None):
        """Return an iterator that yields the data in chunks. Each chunk is a
        dictionary containing numpy arrays with the following keys:

         - ``uvw``: UVW coordinates (position2 - position1), as a Quantity (N×3)
         - ``vis``: visibilities (C×N×P for C channels and P polarizations)
         - ``weights``: imaging weights (C×N×P for C channels and P polarizations)
         - ``baselines``: arbitrary integer baseline IDs; negative IDs indicate autocorrelations
         - ``feed_angle1``: angle between feed and sky (parallactic angle plus a fixed
           offset for the feed), for the first antenna in the baseline (N).
         - ``feed_angle2``: angle between feed and sky for the second antenna in
           the baseline (N).
         - ``progress``: progress made through the file, in some arbitrary units
         - ``total``: size of the file, in same units as ``progress``

        .. note::

           The visibilities are assumed to use a convention in which the phase
           of the electric field *increases* with time, which is consistent with
           [HB1996]_.

           The sign convention for UVW matches the definition for measurement
           sets, but is opposite to that actually used by CASA_ (and thus by
           other imagers that give consistent results with CASA).

           .. [HB1996]

              Understanding radio polarimetry. III. Interpreting the IAU/IEEE definitions
              of the Stokes parameters. J. P. Hamaker and J. D. Bregman. Astron.
              Astrophys. Suppl. Ser., 117 1 (1996) 161-165.

           .. _CASA: http://casa.nrao.edu/Memos/CoordConvention.pdf

        The arrays are indexed first by channel (where applicable) then by a 1D
        time/baseline coordinate. The second index is x/y/z for 'uvw' and
        polarization product for 'vis' and 'weights'.  Flags are not explicitly
        returned: they are either omitted entirely (if all pols are flagged) or
        indicated with a zero weight.

        If :meth:`has_feed_angles` returns ``False``, then `feed_angle1` and
        `feed_angle2` will be absent.

        Parameters
        ----------
        start_channel,stop_channel : int
            Half-open range of channels for which to return data
        max_chunk_vis : int, optional
            Maximum number of full-pol visibilities to return in each chunk.
            If not specified, there is no bound. This is a soft limit that
            may be exceeded if the natural unit of the storage format (e.g. row
            in a measurement set) exceeds this size.
        """
        raise NotImplementedError('Abstract base class')

    def sky_model(self):
        """Get the stored sky model, if any.

        Returns
        -------
        sky_model : :class:`~katsdpimager.sky_model.SkyModel`
            Sky model from the data set

        Raises
        ------
        sky_model.NoSkyModelError
            If there is no sky model in the data set
        """
        return sky_model.NoSkyModelError('This input format does not support sky models')

    def extra_fits_headers(self):
        """Get loader-specific FITS headers to add to the output.

        This is only called after iterating over the data with
        :meth:`data_iter`, so it is possible for :meth:`data_iter` to compute
        data that will be used here. However, note that when :opt:`--vis-limit`
        is specified the data iterator will be closed early.

        Returns
        -------
        headers : astropy.io.fits.Header
            Extra FITS headers to insert into output files. The headers are passed
            to :py:class:`astropy.io.fits.Header`, so for example the value can be
            a (value, comment) tuple.
        """
        return astropy.io.fits.Header()

    @property
    def raw_data(self):
        """Return a handle to the the underlying class-specific data set."""
        raise NotImplementedError('Abstract base class')

    def close(self):
        """Close any open file handles. The object must not be used after this."""
        pass



"""Top-level objects for accelerated imaging, that glue the lower-level
objects together."""

import numpy as np
from katsdpsigproc import accel

from . import grid, predict, weight, image, clean


class ImagingTemplate:
    """Template holding all the other templates for imaging."""

    def __init__(self, command_queue, array_parameters, image_parameters,
                 weight_parameters, grid_parameters, clean_parameters):
        self.command_queue = command_queue
        self.array_parameters = array_parameters
        self.image_parameters = image_parameters
        self.weight_parameters = weight_parameters
        self.grid_parameters = grid_parameters
        self.clean_parameters = clean_parameters
        context = command_queue.context
        num_polarizations = len(image_parameters.polarizations)
        image_shape = (num_polarizations,
                       image_parameters.pixels,
                       image_parameters.pixels)
        # Currently none of the kernels accessing the layer need any padding.
        # It would be nice if there was a cleaner way to handle this; possibly
        # by deferring creation of the FFT plan until instantiation.
        padded_layer_shape = layer_shape = image_shape[1:]
        self.weights = weight.WeightsTemplate(
            context, weight_parameters.weight_type, num_polarizations)
        self.gridder = grid.GridderTemplate(context, image_parameters, grid_parameters)
        self.predict = predict.PredictTemplate(
            context, image_parameters.real_dtype, num_polarizations)
        self.grid_image = image.GridImageTemplate(
            command_queue, layer_shape, padded_layer_shape, image_parameters.real_dtype)
        self.psf_patch = clean.PsfPatchTemplate(
            context, image_parameters.real_dtype, num_polarizations)
        self.noise_est = clean.NoiseEstTemplate(
            context, image_parameters.real_dtype, num_polarizations, clean_parameters.mode)
        self.clean = clean.CleanTemplate(
            context, clean_parameters, image_parameters.real_dtype, num_polarizations)
        self.scale = image.ScaleTemplate(
            context, image_parameters.real_dtype, num_polarizations)
        self.taper1d = accel.SVMArray(
            context, (image_parameters.pixels,), image_parameters.real_dtype)
        self.gridder.convolve_kernel.taper(image_parameters.pixels, self.taper1d)
        if grid_parameters.degrid:
            self.untaper1d = accel.SVMArray(
                context, (image_parameters.pixels,), image_parameters.real_dtype)
            self.degridder = grid.DegridderTemplate(context, image_parameters, grid_parameters)
            self.degridder.convolve_kernel.taper(image_parameters.pixels, self.untaper1d)
        else:
            self.untaper1d = None
            self.degridder = None

    def instantiate(self, *args, **kwargs):
        return Imaging(self, *args, **kwargs)


class Imaging(accel.OperationSequence):
    def __init__(self, template, max_vis, max_sources, major, allocator=None):
        self.template = template
        lm_scale = float(template.image_parameters.pixel_size)
        lm_bias = -0.5 * template.image_parameters.pixels * lm_scale
        image_shape = (len(template.image_parameters.polarizations),
                       template.image_parameters.pixels,
                       template.image_parameters.pixels)
        self._gridder = template.gridder.instantiate(
            template.command_queue, template.array_parameters, max_vis, allocator)
        self._continuum_predict = template.predict.instantiate(
            template.command_queue, template.image_parameters, template.grid_parameters,
            max_vis, max_sources, allocator)
        grid_shape = self._gridder.slots['grid'].shape
        self._weights = template.weights.instantiate(
            template.command_queue, grid_shape, max_vis, allocator)
        self._weights.robustness = template.weight_parameters.robustness
        self._grid_to_image = template.grid_image.instantiate_grid_to_image(
            grid_shape, lm_scale, lm_bias, allocator)
        self._psf_patch = template.psf_patch.instantiate(
            template.command_queue, image_shape, allocator)
        self._noise_est = template.noise_est.instantiate(
            template.command_queue, image_shape, template.clean_parameters.border, allocator)
        self._clean = template.clean.instantiate(
            template.command_queue, template.image_parameters, allocator)
        self._scale = template.scale.instantiate(
            template.command_queue, image_shape, allocator)
        self._grid_to_image.bind(kernel1d=template.taper1d)
        if template.grid_parameters.degrid:
            self._predict = template.degridder.instantiate(
                template.command_queue, template.array_parameters, max_vis, allocator)
            degrid_shape = self._predict.slots['grid'].shape
            self._image_to_grid = template.grid_image.instantiate_image_to_grid(
                degrid_shape, lm_scale, lm_bias, allocator)
            self._image_to_grid.bind(kernel1d=template.untaper1d)
        else:
            max_components = min(template.image_parameters.pixels**2,
                                 (major - 1) * template.clean_parameters.minor)
            self._predict = template.predict.instantiate(
                template.command_queue, template.image_parameters, template.grid_parameters,
                max_vis, max_components, allocator)
        operations = [
            ('weights', self._weights),
            ('gridder', self._gridder),
            ('predict', self._predict),
            ('continuum_predict', self._continuum_predict),
            ('grid_to_image', self._grid_to_image),
            ('psf_patch', self._psf_patch),
            ('noise_est', self._noise_est),
            ('clean', self._clean),
            ('scale', self._scale)
        ]
        compounds = {
            'weights': ['weights:weights'],
            'weights_grid': ['weights:grid', 'gridder:weights_grid'],
            'uv': ['weights:uv', 'gridder:uv', 'predict:uv', 'continuum_predict:uv'],
            'w_plane': ['gridder:w_plane', 'predict:w_plane', 'continuum_predict:w_plane'],
            'vis': ['gridder:vis', 'predict:vis', 'continuum_predict:vis'],
            'predict_weights': ['predict:weights', 'continuum_predict:weights'],
            'grid': ['gridder:grid', 'grid_to_image:grid'],
            'layer': ['grid_to_image:layer'],
            'dirty': ['grid_to_image:image', 'noise_est:dirty', 'clean:dirty', 'scale:data'],
            'model': ['clean:model'],
            'psf': ['clean:psf', 'psf_patch:psf'],
            'tile_max': ['clean:tile_max'],
            'tile_pos': ['clean:tile_pos'],
            'peak_value': ['clean:peak_value'],
            'peak_pos': ['clean:peak_pos'],
            'peak_pixel': ['clean:peak_pixel']
        }
        if template.grid_parameters.degrid:
            operations.append(('image_to_grid', self._image_to_grid))
            compounds['degrid'] = ['predict:grid', 'image_to_grid:grid']
            compounds['layer'].append('image_to_grid:layer')
            compounds['model'].append('image_to_grid:image')
        # TODO: could alias weights with something, since it's only needed
        # early on while setting up weights_grid.
        # TODO: could alias noise_est:rank with something, since it's only
        # needed at the start of the minor cycles (it is small though).
        super().__init__(
            template.command_queue, operations, compounds, allocator=allocator)
        # dirty_to_psf swaps the dirty and PSF images, so they need to have
        # compatible padding.
        for x, y in zip(self.slots['dirty'].dimensions, self.slots['psf'].dimensions):
            x.link(y)

    def __call__(self, **kwargs):
        raise NotImplementedError()

    @property
    def num_vis(self):
        return self._gridder.num_vis

    @num_vis.setter
    def num_vis(self, value):
        self._gridder.num_vis = value
        self._predict.num_vis = value
        self._continuum_predict.num_vis = value

    def clear_weights(self):
        self.ensure_all_bound()
        self._weights.clear()

    def grid_weights(self, uv, weights):
        self.ensure_all_bound()
        self._weights.grid(uv, weights)

    def finalize_weights(self):
        self.ensure_all_bound()
        return self._weights.finalize()

    def clear_grid(self):
        self.ensure_all_bound()
        self.buffer('grid').zero(self.command_queue)

    def clear_dirty(self):
        self.ensure_all_bound()
        self.buffer('dirty').zero(self.command_queue)

    def clear_model(self):
        self.ensure_all_bound()
        self.buffer('model').zero(self.command_queue)

    def set_coordinates(self, *args, **kwargs):
        self.ensure_all_bound()
        # The gridder and predictors share their coordinates, so we can
        # use any here.
        self._gridder.set_coordinates(*args, **kwargs)

    def set_vis(self, *args, **kwargs):
        self.ensure_all_bound()
        # The gridder and predicters share their visibilities, so we
        # can use any here.
        self._gridder.set_vis(*args, **kwargs)

    def set_weights(self, *args, **kwargs):
        """Set statistical weights for prediction"""
        self.ensure_all_bound()
        # The predicters share their weights, so we can use any here.
        self._predict.set_weights(*args, **kwargs)

    def grid(self):
        self.ensure_all_bound()
        self._gridder()

    def predict(self, w):
        self.ensure_all_bound()
        if not self.template.grid_parameters.degrid:
            self._predict.set_w(w)
        self._predict()

    def continuum_predict(self, w):
        self.ensure_all_bound()
        self._continuum_predict.set_w(w)
        self._continuum_predict()

    def set_sky_model(self, sky_model, phase_centre):
        self.ensure_all_bound()
        self._continuum_predict.set_sky_model(sky_model, phase_centre)

    def grid_to_image(self, w):
        self.ensure_all_bound()
        self._grid_to_image.set_w(w)
        self._grid_to_image()

    def model_to_grid(self, w):
        if not self._image_to_grid:
            raise RuntimeError('Can only use model_to_grid with degridding')
        self.ensure_all_bound()
        self._image_to_grid.set_w(w)
        self._image_to_grid()

    def model_to_predict(self):
        if self.template.grid_parameters.degrid:
            raise RuntimeError('Can only use model_to_predict with direct prediction')
        # It's computed on the CPU, so we need to be sure that device work is complete
        self._predict.command_queue.finish()
        self._predict.set_sky_image(self.buffer('model'))

    def scale_dirty(self, scale_factor):
        self.ensure_all_bound()
        self._scale.set_scale_factor(scale_factor)
        self._scale()

    def dirty_to_psf(self):
        dirty = self.buffer('dirty')
        psf = self.buffer('psf')
        self.bind(dirty=psf, psf=dirty)

    def psf_patch(self):
        self.ensure_all_bound()
        return self._psf_patch(self.template.clean_parameters.psf_cutoff,
                               self.template.clean_parameters.psf_limit)

    def noise_est(self):
        self.ensure_all_bound()
        return self._noise_est()

    def clean_reset(self):
        self.ensure_all_bound()
        self._clean.reset()

    def clean_cycle(self, psf_patch, threshold=0.0):
        self.ensure_all_bound()
        return self._clean(psf_patch, threshold)


class ImagingHost:
    """Host-only equivalent to :class:`Imaging`."""

    def __init__(self, image_parameters, weight_parameters, grid_parameters, clean_parameters):
        lm_scale = float(image_parameters.pixel_size)
        lm_bias = -0.5 * image_parameters.pixels * lm_scale
        self._clean_parameters = clean_parameters
        self._image_parameters = image_parameters
        self._gridder = grid.GridderHost(image_parameters, grid_parameters)
        self._grid = self._gridder.values
        self._continuum_predict = predict.PredictHost(image_parameters, grid_parameters)
        self._weights_grid = self._gridder.weights_grid
        self._weights = weight.WeightsHost(weight_parameters.weight_type, self._weights_grid)
        self._weights.robustness = weight_parameters.robustness
        self._layer = np.empty(self._grid.shape, image_parameters.complex_dtype)
        self._dirty = np.empty(self._grid.shape, image_parameters.real_dtype)
        self._model = np.empty(self._grid.shape, image_parameters.real_dtype)
        self._psf = np.empty(self._grid.shape, image_parameters.real_dtype)
        self._grid_to_image = image.GridToImageHost(
            self._grid, self._layer, self._dirty,
            self._gridder.kernel.taper(image_parameters.pixels), lm_scale, lm_bias)
        self._clean = clean.CleanHost(image_parameters, clean_parameters,
                                      self._dirty, self._psf, self._model)
        if grid_parameters.degrid:
            self._predict = grid.DegridderHost(image_parameters, grid_parameters)
            self._degrid = self._predict.values
            self._image_to_grid = image.ImageToGridHost(
                self._degrid, self._layer, self._model,
                self._predict.kernel.taper(image_parameters.pixels), lm_scale, lm_bias)
        else:
            self._predict = predict.PredictHost(image_parameters, grid_parameters)
            self._degrid = None
            self._image_to_grid = None
        self._buffer = {
            'psf': self._psf,
            'dirty': self._dirty,
            'model': self._model,
            'grid': self._grid,
            'weights_grid': self._weights_grid
        }
        if self._degrid is not None:
            self._buffer['degrid'] = self._degrid

    def buffer(self, name):
        return self._buffer[name]

    @property
    def num_vis(self):
        return self._gridder.num_vis

    @num_vis.setter
    def num_vis(self, value):
        self._gridder.num_vis = value
        self._predict.num_vis = value
        self._continuum_predict.num_vis = value

    def clear_weights(self):
        self._weights_grid.fill(0)

    def grid_weights(self, uv, weights):
        self._weights.grid(uv, weights)

    def finalize_weights(self):
        return self._weights.finalize()

    def clear_grid(self):
        self._grid.fill(0)

    def clear_dirty(self):
        self._dirty.fill(0)

    def clear_model(self):
        self._model.fill(0)

    def set_sky_model(self, sky_model, phase_centre):
        self._continuum_predict.set_sky_model(sky_model, phase_centre)

    def set_coordinates(self, *args, **kwargs):
        self._gridder.set_coordinates(*args, **kwargs)
        self._predict.set_coordinates(*args, **kwargs)
        self._continuum_predict.set_coordinates(*args, **kwargs)

    def set_vis(self, *args, **kwargs):
        self._gridder.set_vis(*args, **kwargs)
        self._predict.set_vis(*args, **kwargs)
        self._continuum_predict.set_vis(*args, **kwargs)

    def set_weights(self, *args, **kwargs):
        self._predict.set_weights(*args, **kwargs)
        self._continuum_predict.set_weights(*args, **kwargs)

    def grid(self):
        self._gridder()

    def predict(self, w):
        if self._degrid is None:
            self._predict.set_w(w)
        self._predict()

    def continuum_predict(self, w):
        self._continuum_predict.set_w(w)
        self._continuum_predict()

    def grid_to_image(self, w):
        self._grid_to_image.set_w(w)
        self._grid_to_image()

    def model_to_grid(self, w):
        if self._degrid is None:
            raise RuntimeError('Can only use model_to_grid with degridding')
        self._image_to_grid.set_w(w)
        self._image_to_grid()

    def model_to_predict(self):
        if self._degrid is not None:
            raise RuntimeError('Can only use model_to_predict with direct prediction')
        self._predict.set_sky_image(self._model)

    def scale_dirty(self, scale_factor):
        self._dirty *= scale_factor[:, np.newaxis, np.newaxis]

    def dirty_to_psf(self):
        self._psf[:] = self._dirty

    def psf_patch(self):
        return clean.psf_patch_host(self._psf,
                                    self._clean_parameters.psf_cutoff,
                                    self._clean_parameters.psf_limit)

    def noise_est(self):
        return clean.noise_est_host(self._dirty, self._clean_parameters.border,
                                    self._clean_parameters.mode)

    def clean_reset(self):
        self._clean.reset()

    def clean_cycle(self, psf_patch, threshold=0.0):
        return self._clean(psf_patch, threshold)



"""Tests for :mod:`katsdpimager.fft`"""

import numpy as np
import katsdpsigproc.accel as accel
from katsdpsigproc.test.test_accel import device_test, cuda_test

from .. import fft
from .utils import RandomState


class TestFftshift:
    @classmethod
    def pad_dimension(cls, dim, extra):
        """Modifies `dim` to have at least `extra` padding"""
        newdim = accel.Dimension(dim.size, min_padded_size=dim.size + extra)
        newdim.link(dim)

    @device_test
    def test_2d(self, context, command_queue):
        template = fft.FftshiftTemplate(context, np.int32)
        fn = template.instantiate(command_queue, (10, 6))
        fn.ensure_all_bound()
        data = fn.buffer('data')
        host_data = (np.arange(6).reshape(1, 6) + np.arange(10).reshape(10, 1)).astype(np.int32)
        data.set(command_queue, host_data)
        expected = np.fft.fftshift(host_data)
        fn()
        actual = data.get(command_queue)
        np.testing.assert_equal(expected, actual)

    @device_test
    def test_3d(self, context, command_queue):
        template = fft.FftshiftTemplate(context, np.int32)
        fn = template.instantiate(command_queue, (3, 10, 6))
        # Uses padded data to detect padding bugs
        self.pad_dimension(fn.slots['data'].dimensions[0], 1)
        self.pad_dimension(fn.slots['data'].dimensions[1], 3)
        self.pad_dimension(fn.slots['data'].dimensions[2], 5)
        fn.ensure_all_bound()
        data = fn.buffer('data')
        host_data = np.arange(3 * 10 * 6).reshape(3, 10, 6).astype(np.int32)
        data.set(command_queue, host_data)
        expected = np.fft.fftshift(host_data, axes=(1, 2))
        fn()
        actual = data.get(command_queue)
        np.testing.assert_equal(expected, actual)


class TestFft:
    @device_test
    @cuda_test
    def test_forward(self, context, command_queue):
        rs = RandomState(1)
        template = fft.FftTemplate(
            command_queue, 2, (3, 2, 16, 48), np.complex64, np.complex64,
            (4, 5, 24, 64), (4, 5, 20, 48))
        fn = template.instantiate(fft.FFT_FORWARD, allocator=accel.SVMAllocator(context))
        fn.ensure_all_bound()
        src = fn.buffer('src')
        dest = fn.buffer('dest')
        src[:] = rs.complex_normal(size=src.shape).astype(np.complex64)
        fn()
        command_queue.finish()
        expected = np.fft.fftn(src, axes=(2, 3))
        np.testing.assert_allclose(expected, dest, rtol=1e-4)

    def _test_r2c(self, context, command_queue, N, shape, padded_shape_src, padded_shape_dest):
        rs = np.random.RandomState(1)
        template = fft.FftTemplate(
            command_queue, N, shape, np.float32, np.complex64, padded_shape_src, padded_shape_dest)
        fn = template.instantiate(fft.FFT_FORWARD, allocator=accel.SVMAllocator(context))
        fn.ensure_all_bound()
        src = fn.buffer('src')
        dest = fn.buffer('dest')
        src[:] = rs.standard_normal(src.shape)
        fn()
        command_queue.finish()
        expected = np.fft.rfftn(src, axes=(2, 3))
        np.testing.assert_allclose(expected, dest, rtol=1e-4)

    def _test_c2r(self, context, command_queue, N, shape, padded_shape_src, padded_shape_dest):
        rs = np.random.RandomState(1)
        template = fft.FftTemplate(
            command_queue, N, shape, np.complex64, np.float32, padded_shape_src, padded_shape_dest)
        fn = template.instantiate(fft.FFT_INVERSE, allocator=accel.SVMAllocator(context))
        fn.ensure_all_bound()
        src = fn.buffer('src')
        dest = fn.buffer('dest')
        signal = rs.standard_normal(shape) + 3
        src[:] = np.fft.rfftn(signal, axes=(2, 3))
        fn()
        command_queue.finish()
        expected = signal * shape[2] * shape[3]
        np.testing.assert_allclose(expected, dest, rtol=1e-4)

    @device_test
    @cuda_test
    def test_r2c_even(self, context, command_queue):
        self._test_r2c(context, command_queue, 2, (3, 2, 16, 48), (4, 5, 24, 64), (4, 5, 20, 27))

    @device_test
    @cuda_test
    def test_r2c_odd(self, context, command_queue):
        self._test_r2c(context, command_queue, 2, (3, 2, 15, 47), (4, 5, 23, 63), (4, 5, 17, 24))

    @device_test
    @cuda_test
    def test_c2r_even(self, context, command_queue):
        self._test_c2r(context, command_queue, 2, (3, 2, 16, 48), (4, 5, 20, 27), (4, 5, 24, 64))

    @device_test
    @cuda_test
    def test_c2r_odd(self, context, command_queue):
        self._test_c2r(context, command_queue, 2, (3, 2, 15, 47), (4, 5, 17, 24), (4, 5, 23, 63))

    @device_test
    @cuda_test
    def test_inverse(self, context, command_queue):
        rs = RandomState(1)
        template = fft.FftTemplate(
            command_queue, 2, (3, 2, 16, 48), np.complex64, np.complex64,
            (4, 5, 24, 64), (4, 5, 20, 48))
        fn = template.instantiate(fft.FFT_INVERSE, allocator=accel.SVMAllocator(context))
        fn.ensure_all_bound()
        src = fn.buffer('src')
        dest = fn.buffer('dest')
        src[:] = rs.complex_normal(size=src.shape).astype(np.complex64)
        fn()
        command_queue.finish()
        expected = np.fft.ifftn(src, axes=(2, 3)) * (src.shape[2] * src.shape[3])
        np.testing.assert_allclose(expected, dest, rtol=1e-4)



"""Tests for :mod:`katsdpimager.predict`"""

import numpy as np
import katpoint
import katsdpsigproc.accel as accel
from katsdpsigproc.test.test_accel import device_test, force_autotune
from astropy import units
from nose.tools import assert_equal

from .. import predict, parameters, polarization, sky_model
from .utils import RandomState


class TestPredict:
    """Tests for :class:`.Predict`"""

    def setup(self):
        # Deliberately use a strange subset of polarizations, to test that
        # the indexing is correct.
        pols = [polarization.STOKES_I,
                polarization.STOKES_Q,
                polarization.STOKES_V]
        self.image_parameters = parameters.ImageParameters(
            q_fov=1.0,
            image_oversample=None,
            frequency=0.2 * units.m,
            array=None,
            polarizations=pols,
            dtype=np.float64,
            pixel_size=0.00001,
            pixels=4096)
        oversample = 8
        w_planes = 100
        self.grid_parameters = parameters.GridParameters(
            antialias_width=7.0,
            oversample=oversample,
            image_oversample=4,
            w_slices=10,
            w_planes=w_planes,
            max_w=5 * units.m,
            kernel_width=7)
        catalogue = katpoint.Catalogue([
            "dummy0, radec, 19:39:25.03, -63:42:45.7, (200.0 12000.0 -11.11 7.777 -1.231 0 0 0 1 0.1 0 0)",       # noqa: E501
            "dummy1, radec, 19:39:20.38, -63:42:09.1, (800.0 8400.0 -3.708 3.807 -0.7202 0 0 0 1 0.2 0.2 0.2)",   # noqa: E501
            "dummy2, radec, 19:39:08.29, -63:42:33.0, (800.0 43200.0 0.956 0.584 -0.1644 0 0 0 1 0.1 0 1)"        # noqa: E501
        ])
        self.model = sky_model.KatpointSkyModel(catalogue)
        self.phase_centre = katpoint.construct_radec_target(
            '19:39:30', '-63:42:30').astrometric_radec()
        self.phase_centre = self.phase_centre * units.rad

    def _test_random(self, context, queue, n_vis):
        ip = self.image_parameters
        gp = self.grid_parameters
        rs = RandomState(seed=1)
        uv = rs.random_integers(-2048, 2048, size=(n_vis, 2)).astype(np.int16)
        sub_uv = rs.random_integers(0, gp.oversample - 1, size=(n_vis, 2)).astype(np.int16)
        w_plane = rs.random_integers(0, gp.w_planes - 1, size=n_vis).astype(np.int16)
        weights = rs.uniform(size=(n_vis, len(ip.polarizations))).astype(np.float32)
        vis = rs.complex_normal(size=(n_vis, len(ip.polarizations)))

        allocator = accel.SVMAllocator(context)
        template = predict.PredictTemplate(context, np.float32, len(ip.polarizations))
        fn = template.instantiate(queue, ip, gp,
                                  n_vis, len(self.model), allocator=allocator)
        fn.ensure_all_bound()
        fn.num_vis = n_vis
        fn.set_coordinates(uv, sub_uv, w_plane)
        fn.set_vis(vis)
        fn.set_weights(weights)
        fn.set_sky_model(self.model, self.phase_centre)
        fn.set_w(1.2)
        fn()

        host = predict.PredictHost(ip, gp)
        host.num_vis = n_vis
        host.set_coordinates(uv, sub_uv, w_plane)
        host.set_vis(vis)
        host.set_weights(weights)
        host.set_sky_model(self.model, self.phase_centre)
        host.set_w(1.2)
        host()

        queue.finish()
        device_vis = fn.buffer('vis')[:n_vis]
        # Accuracy of individual visibilities is low, because the delay can have
        # a large whole number of wavelengths which degrades the precision of the
        # phase. Differences in where FMAs are inserted by compilers causes
        # differences.
        np.testing.assert_allclose(device_vis, vis, rtol=5e-4)

    @device_test
    def test_random(self, context, queue):
        """Compare the host and device versions with random data"""
        self._test_random(context, queue, 1001)

    @device_test
    def test_few_vis(self, context, queue):
        """Test with fewer visibilities than sources.

        This is a regression test: the first version had a bug in this case.
        """
        self._test_random(context, queue, 2)

    def test_extract_sky_image(self):
        ip = self.image_parameters
        image = np.zeros((len(ip.polarizations), ip.pixels, ip.pixels), np.float32)
        # Note: first index is m, not l
        image[:, 2048, 2048] = [1, 2, 3]
        image[:, 1024, 512] = [2.5, 1.5, 0.0]
        image[:, 0, 4095] = [4, 0, 0]
        image[:, 4095, 0] = [5, 1, 2]
        lmn, flux = predict._extract_sky_image(self.image_parameters, self.grid_parameters, image)
        assert_equal(lmn.shape, (4, 3))
        assert_equal(flux.shape, (4, 3))
        np.testing.assert_allclose(lmn[:, 0:2], [
            [2047e-5, -2048e-5],
            [-1536e-5, -1024e-5],
            [0, 0],
            [-2048e-5, 2047e-5]
        ])
        expected_flux = np.array([
            [4.0, 0, 0],
            [2.5, 1.5, 0.0],
            [1, 2, 3],
            [5, 1, 2]
        ])
        expected_flux[0] *= np.sinc(0.5 / 8) * np.sinc(2047 / 4096 / 8)
        expected_flux[1] *= np.sinc(0.25 / 8) * np.sinc(0.375 / 8)
        expected_flux[3] *= np.sinc(2047 / 4096 / 8) * np.sinc(0.5 / 8)
        np.testing.assert_allclose(flux, expected_flux)

    @device_test
    @force_autotune
    def test_autotune(self, context, queue):
        """Check that autotuner runs successfully"""
        predict.PredictTemplate(context, np.float32, 4)
        predict.PredictTemplate(context, np.float64, 4)



"""Progress bar functions.

This wraps the :py:mod:`progress` module to add certain enhancements:

- A helper is provided for steps where no partial progress will be reported, but
  goes from empty to full in one step.
- A fallback implementation is used when stderr is not a TTY.
"""

import sys
import functools
from contextlib import contextmanager

import progress
import progress.bar


class NoTtyMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._written_length = 0

    def update(self):
        filled_length = int(self.width * self.progress)
        message = self.message % self
        bar = self.fill * filled_length
        line = ''.join([message, self.bar_prefix, bar])
        if len(line) > self._written_length:
            self.file.write(line[self._written_length:])
            self.file.flush()
            self._written_length = len(line)

    def finish(self):
        filled_length = int(self.width * self.progress)
        empty_length = self.width - filled_length
        empty = self.empty_fill * empty_length
        suffix = self.suffix % self
        self.file.write(''.join([empty, self.bar_suffix, suffix, '\n']))
        self.file.flush()


class NoTtyBar(NoTtyMixin, progress.bar.Bar):
    pass


def make_progressbar(name, *args, **kwargs):
    if sys.stderr.isatty():
        bar = progress.bar.Bar("{:20}".format(name),
                               suffix='%(percent)3d%% [%(eta_td)s]',
                               *args, **kwargs)
    else:
        bar = NoTtyBar("{:20}".format(name), suffix='', *args, **kwargs)
    bar.update()
    return bar


class RunOnce:
    """Wrapper callable that calls the wrapped function only the first time."""
    def __init__(self, wrapped):
        self.wrapped = wrapped
        self.run = False

    def __call__(self, *args, **kwargs):
        if not self.run:
            self.run = True
            self.wrapped(*args, **kwargs)


def run_once(func):
    """Decorator that modifies a function to only run the first time."""
    wrapper = RunOnce(func)
    functools.update_wrapper(wrapper, func)
    return wrapper


@contextmanager
def finishing(bar):
    """Context manager that ensures that a progress bar is always finished,
    even if an exception occurs. This avoids cases where the cursor is hidden
    in the terminal after hitting Ctrl-C."""
    bar.finish = run_once(bar.finish)
    yield bar
    bar.finish()


@contextmanager
def step(name):
    bar = make_progressbar(name, max=1)
    yield
    bar.next()
    bar.finish()



import tempfile
import os
import atexit
import logging
from abc import abstractmethod

import numpy as np
from astropy import units
import katsdpsigproc.accel as accel

from . import \
    loader, parameters, polarization, preprocess, clean, weight, sky_model, \
    imaging, progress, beam


logger = logging.getLogger(__name__)


def parse_quantity(str_value):
    """Parse a string into an astropy Quantity. Rather than trying to guess
    where the split occurs, we try every position from the back until we
    succeed."""
    for i in range(len(str_value), 0, -1):
        try:
            value = float(str_value[:i])
            unit = units.Unit(str_value[i:])
            return units.Quantity(value, unit)
        except ValueError:
            pass
    raise ValueError('Could not parse {} as a quantity'.format(str_value))


def preprocess_visibilities(dataset, args, start_channel, stop_channel,
                            image_parameters, grid_parameters, polarization_matrices):
    bar = None
    if args.tmp_file:
        handle, filename = tempfile.mkstemp('.h5')
        os.close(handle)
        atexit.register(os.remove, filename)
        collector = preprocess.VisibilityCollectorHDF5(
            filename, image_parameters, grid_parameters, args.vis_block,
            max_cache_size=args.max_cache_size)
    else:
        collector = preprocess.VisibilityCollectorMem(
            image_parameters, grid_parameters, args.vis_block)
    try:
        for chunk in loader.data_iter(dataset, args.vis_limit, args.vis_load,
                                      start_channel, stop_channel):
            if bar is None:
                bar = progress.make_progressbar("Preprocessing vis", max=chunk['total'])
            collector.add(
                chunk['uvw'], chunk['weights'], chunk['baselines'], chunk['vis'],
                chunk.get('feed_angle1'), chunk.get('feed_angle2'),
                *polarization_matrices)
            bar.goto(chunk['progress'])
    finally:
        if bar is not None:
            bar.finish()
        collector.close()
    logger.info("Compressed %d visibilities to %d (%.2f%%)",
                collector.num_input, collector.num_output,
                100.0 * collector.num_output / collector.num_input)
    return collector


def make_weights(queue, reader, rel_channel, imager, weight_type, vis_block):
    imager.clear_weights()
    total = 0
    for w_slice in range(reader.num_w_slices(rel_channel)):
        total += reader.len(rel_channel, w_slice)
    bar = progress.make_progressbar('Computing weights', max=total)
    queue.finish()
    with progress.finishing(bar):
        if weight_type != weight.WeightType.NATURAL:
            for w_slice in range(reader.num_w_slices(rel_channel)):
                for chunk in reader.iter_slice(rel_channel, w_slice, vis_block):
                    imager.grid_weights(chunk.uv, chunk.weights)
                    # Need to serialise calls to grid, since otherwise the next
                    # call will overwrite the incoming data before the previous
                    # iteration is done with it.
                    queue.finish()
                    bar.next(len(chunk.uv))
        else:
            bar.next(total)
        normalized_rms = imager.finalize_weights()
        queue.finish()
    logger.info('Normalized thermal RMS: %g', normalized_rms)


def make_dirty(queue, reader, rel_channel, name, field, imager, mid_w, vis_block, degrid,
               full_cycle=False, subtract_model=None):
    imager.clear_dirty()
    queue.finish()
    if full_cycle and not degrid:
        with progress.step('Extract components'):
            imager.model_to_predict()
    for w_slice in range(reader.num_w_slices(rel_channel)):
        N = reader.len(rel_channel, w_slice)
        if N == 0:
            logger.info("Skipping slice %d which has no visibilities", w_slice + 1)
            continue
        label = '{} {}/{}'.format(name, w_slice + 1, reader.num_w_slices(rel_channel))
        if full_cycle and degrid:
            with progress.step('FFT {}'.format(label)):
                imager.model_to_grid(mid_w[w_slice])
        bar = progress.make_progressbar('Grid {}'.format(label), max=N)
        imager.clear_grid()
        queue.finish()
        with progress.finishing(bar):
            for chunk in reader.iter_slice(rel_channel, w_slice, vis_block):
                imager.num_vis = len(chunk.uv)
                imager.set_coordinates(chunk.uv, chunk.sub_uv, chunk.w_plane)
                imager.set_vis(chunk[field])
                if full_cycle or subtract_model:
                    imager.set_weights(chunk.weights)
                if subtract_model:
                    imager.continuum_predict(mid_w[w_slice])
                if full_cycle:
                    imager.predict(mid_w[w_slice])
                imager.grid()
                # Need to serialise calls to grid, since otherwise the next
                # call will overwrite the incoming data before the previous
                # iteration is done with it.
                queue.finish()
                bar.next(len(chunk))

        with progress.step('IFFT {}'.format(label)):
            imager.grid_to_image(mid_w[w_slice])
            queue.finish()


def extract_psf(psf, psf_patch):
    """Extract the central region of a PSF.

    This function is 2D: the polarization axis must have already been removed.
    """
    y0 = (psf.shape[0] - psf_patch[0]) // 2
    y1 = y0 + psf_patch[0]
    x0 = (psf.shape[1] - psf_patch[1]) // 2
    x1 = x0 + psf_patch[1]
    return psf[..., y0:y1, x0:x1]


def log_parameters(name, params):
    if logger.isEnabledFor(logging.INFO):
        s = str(params)
        lines = s.split('\n')
        logger.info('%s:', name)
        for line in lines:
            if line:
                logger.info('    %s', line)


class ChannelParameters:
    """Collects imaging parameters for a single channel.

    Parameters
    ----------
    args : :class:`argparse.Namespace`
        Command-line arguments
    dataset : :class:`loader_core.LoaderBase`
        Input dataset
    channel : int
        Index of this channel in the dataset
    array_p : :class:`katsdpimager.parameters.ArrayParameters`
        Array parameters from the input file

    Attributes
    ----------
    channel : int
        Index of this channel in the dataset
    image_p : :class:`katsdpimager.parameters.ImageParameters`
        Image parameters
    grid_p : :class:`katsdpimager.parameters.GridParameters`
        Gridding parameters
    clean_p : :class:`katsdpimage.parameters.CleanParameters`
        CLEAN parameters
    """

    def __init__(self, args, dataset, channel, array_p):
        self.channel = channel
        self.image_p = parameters.ImageParameters(
            args.q_fov, args.image_oversample,
            dataset.frequency(channel), array_p, args.stokes,
            (np.float32 if args.precision == 'single' else np.float64),
            args.pixel_size, args.pixels)
        if args.max_w is None:
            max_w = array_p.longest_baseline
        elif args.max_w.unit.physical_type == 'dimensionless':
            max_w = args.max_w * self.image_p.wavelength
        else:
            max_w = args.max_w
        if args.w_slices is None:
            w_slices = parameters.w_slices(self.image_p, max_w, args.eps_w,
                                           args.kernel_width, args.aa_width)
        else:
            w_slices = args.w_slices
        if args.w_step.unit.physical_type == 'length':
            w_planes = float(max_w / args.w_step)
        elif args.w_step.unit.physical_type == 'dimensionless':
            w_step = args.w_step * self.image_p.cell_size / args.grid_oversample
            w_planes = float(max_w / w_step)
        else:
            raise ValueError('--w-step must be dimensionless or a length')
        w_planes = int(np.ceil(w_planes / w_slices))
        self.grid_p = parameters.GridParameters(
            args.aa_width, args.grid_oversample, args.kernel_image_oversample,
            w_slices, w_planes, max_w, args.kernel_width, args.degrid)
        if args.clean_mode == 'I':
            clean_mode = clean.CLEAN_I
        elif args.clean_mode == 'IQUV':
            clean_mode = clean.CLEAN_SUMSQ
        else:
            raise ValueError('Unhandled --clean-mode {}'.format(args.clean_mode))
        border = int(round(self.image_p.pixels * args.border))
        limit = int(round(self.image_p.pixels * args.psf_limit))
        self.clean_p = parameters.CleanParameters(
            args.minor, args.loop_gain, args.major_gain, args.threshold,
            clean_mode, args.psf_cutoff, limit, border)

    def log_parameters(self, suffix=''):
        log_parameters("Image parameters" + suffix, self.image_p)
        log_parameters("Grid parameters" + suffix, self.grid_p)
        log_parameters("CLEAN parameters" + suffix, self.clean_p)


def add_options(parser):
    group = parser.add_argument_group('Input selection')
    group.add_argument('--input-option', '-i', action='append', default=[], metavar='KEY=VALUE',
                       help='Backend-specific input parsing option')
    group.add_argument('--start-channel', '-c', type=int, default=0,
                       help='Index of first channel to process [%(default)s]')
    group.add_argument('--stop-channel', '-C', type=int,
                       help='Index past last channel to process [#channels]')
    group.add_argument('--subtract', metavar='URL',
                       help='Sky model with sources to subtract at the start')

    group = parser.add_argument_group('Image options')
    group.add_argument('--q-fov', type=float, default=1.0,
                       help='Field of view to image, relative to main lobe of beam [%(default)s]')
    group.add_argument('--image-oversample', type=float, default=5,
                       help='Pixels per beam [%(default)s]')
    group.add_argument('--pixel-size', type=parse_quantity,
                       help='Size of each image pixel [computed from array]')
    group.add_argument('--pixels', type=int,
                       help='Number of pixels in image [computed from array]')
    group.add_argument('--stokes', type=polarization.parse_stokes, default='I',
                       help='Stokes parameters to image e.g. IQUV for full-Stokes [%(default)s]')
    group.add_argument('--precision', choices=['single', 'double'], default='single',
                       help='Internal floating-point precision [%(default)s]')

    group = parser.add_argument_group('Weighting options')
    group.add_argument('--weight-type',
                       choices=[weight_type.name.lower() for weight_type in weight.WeightType],
                       default='natural',
                       help='Imaging density weights [%(default)s]')
    group.add_argument('--robustness', type=float, default=0.0,
                       help='Robustness parameter for --weight-type=robust [%(default)s]')

    group = parser.add_argument_group('Gridding options')
    group.add_argument('--grid-oversample', type=int, default=8,
                       help='Oversampling factor for convolution kernels [%(default)s]')
    group.add_argument('--kernel-image-oversample', type=int, default=4,
                       help='Oversampling factor for kernel generation [%(default)s]')
    group.add_argument('--w-slices', type=int,
                       help='Number of W slices [computed from --kernel-width]')
    group.add_argument('--w-step', type=parse_quantity, default='1.0',
                       help='Separation between W planes, in subgrid cells or a distance '
                            '[%(default)s]')
    group.add_argument('--max-w', type=parse_quantity,
                       help='Largest w, as either distance or wavelengths [longest baseline]')
    group.add_argument('--aa-width', type=float, default=7,
                       help='Support of anti-aliasing kernel [%(default)s]')
    group.add_argument('--kernel-width', type=int, default=60,
                       help='Support of combined anti-aliasing + w kernel [computed]')
    group.add_argument('--eps-w', type=float, default=0.001,
                       help='Level at which to truncate W kernel [%(default)s]')
    group.add_argument('--degrid', action='store_true',
                       help='Use degridding rather than direct prediction (less accurate)')

    group = parser.add_argument_group('Cleaning options')
    group.add_argument('--psf-cutoff', type=float, default=0.01,
                       help='fraction of PSF peak at which to truncate PSF [%(default)s]')
    group.add_argument('--psf-limit', type=float, default=0.5,
                       help='maximum fraction of image to use for PSF [%(default)s]')
    group.add_argument('--loop-gain', type=float, default=0.1,
                       help='Loop gain for cleaning [%(default)s]')
    group.add_argument('--major-gain', type=float, default=0.85,
                       help='Fraction of peak to clean in each major cycle [%(default)s]')
    group.add_argument('--threshold', type=float, default=5.0,
                       help='CLEAN threshold in sigma [%(default)s]')
    group.add_argument('--major', type=int, default=1,
                       help='Major cycles [%(default)s]')
    group.add_argument('--minor', type=int, default=10000,
                       help='Max minor cycles per major cycle [%(default)s]')
    group.add_argument('--border', type=float, default=0.02,
                       help='CLEAN border as a fraction of image size [%(default)s]')
    group.add_argument('--clean-mode', choices=['I', 'IQUV'], default='IQUV',
                       help='Stokes parameters to consider for peak-finding [%(default)s]')

    group = parser.add_argument_group('Performance tuning options')
    group.add_argument('--vis-block', type=int, default=1048576,
                       help='Number of visibilities to load and grid at a time [%(default)s]')
    group.add_argument('--vis-load', type=int, default=32 * 1048576,
                       help='Number of visibilities to load from file at a time [%(default)s]')
    group.add_argument('--channel-batch', type=int, default=16,
                       help='Number of channels to fully process before starting next batch '
                            '[%(default)s]')
    group.add_argument('--no-tmp-file', dest='tmp_file', action='store_false', default=True,
                       help='Keep preprocessed visibilities in memory')
    group.add_argument('--max-cache-size', type=int, default=None,
                       help='Limit HDF5 cache size for preprocessing')


class Writer:
    """Abstract class that handles writing grids/images to files"""
    @abstractmethod
    def write_fits_image(self, name, description, dataset, image, image_parameters, channel,
                         beam=None, bunit='JY/BEAM'):
        """Optionally output a FITS image.

        `name` is a machine-readable name (consistent with the argument naming
        in image.py, while `description` is human-readable. The other
        arguments have the same meaning as for :meth:`.io.write_fits_image`.
        """

    @abstractmethod
    def write_fits_grid(self, name, description, fftshift, grid_data, image_parameters, channel):
        """Optional output a FITS image showing the UV plane.

        `name` is a machine-readable name (consistent with the argument naming
        in image.py, while `description` is human-readable. If `fftshift` is
        true, the data needs to be fft-shifted on the u and v axes. The other
        arguments have the same meaning as for :meth:`.io.write_fits_grid`.
        """


def process_channel(dataset, args, start_channel,
                    context, queue, reader, writer, channel_p, array_p, weight_p,
                    subtract_model):
    channel = channel_p.channel
    rel_channel = channel - start_channel
    image_p = channel_p.image_p
    grid_p = channel_p.grid_p
    clean_p = channel_p.clean_p

    # Check if there is anything to do
    if not any(reader.len(rel_channel, w_slice)
               for w_slice in range(reader.num_w_slices(rel_channel))):
        logger.info('Skipping channel %d which has no data', channel)
        return

    logger.info('Processing channel %d', channel)
    # Create data and operation instances
    if args.host:
        imager = imaging.ImagingHost(image_p, weight_p, grid_p, clean_p)
    else:
        allocator = accel.SVMAllocator(context)
        imager_template = imaging.ImagingTemplate(
            queue, array_p, image_p, weight_p, grid_p, clean_p)
        n_sources = len(subtract_model) if subtract_model else 0
        imager = imager_template.instantiate(args.vis_block, n_sources, args.major, allocator)
        imager.ensure_all_bound()
    psf = imager.buffer('psf')
    dirty = imager.buffer('dirty')
    model = imager.buffer('model')
    grid_data = imager.buffer('grid')
    imager.clear_model()

    # Compute imaging weights
    make_weights(queue, reader, rel_channel,
                 imager, weight_p.weight_type, args.vis_block)
    writer.write_fits_image('weights', 'image weights',
                            dataset, imager.buffer('weights_grid'), image_p, channel, bunit=None)

    # Create PSF
    slice_w_step = float(grid_p.max_w / image_p.wavelength / (grid_p.w_slices - 0.5))
    mid_w = np.arange(grid_p.w_slices) * slice_w_step
    make_dirty(queue, reader, rel_channel,
               'PSF', 'weights', imager, mid_w, args.vis_block, args.degrid)
    # Normalization
    psf_peak = dirty[..., dirty.shape[1] // 2, dirty.shape[2] // 2]
    if np.any(psf_peak == 0):
        logger.info('Skipping channel %d which has no usable data', channel)
        return
    scale = np.reciprocal(psf_peak)
    imager.scale_dirty(scale)
    queue.finish()
    imager.dirty_to_psf()
    # dirty_to_psf works by swapping, so re-fetch the buffer pointers
    psf = imager.buffer('psf')
    dirty = imager.buffer('dirty')
    psf_patch = imager.psf_patch()
    logger.info('Using %dx%d patch for PSF', psf_patch[2], psf_patch[1])
    # Extract the patch for beam fitting
    psf_core = extract_psf(psf[0], psf_patch[1:])
    restoring_beam = beam.fit_beam(psf_core)
    writer.write_fits_image('psf', 'PSF', dataset, psf, image_p, channel, restoring_beam)

    # Imaging
    if subtract_model:
        imager.set_sky_model(subtract_model, dataset.phase_centre())
    for i in range(args.major):
        logger.info("Starting major cycle %d/%d", i + 1, args.major)
        make_dirty(queue, reader, rel_channel,
                   'image', 'vis', imager, mid_w, args.vis_block, args.degrid,
                   i != 0, subtract_model)
        imager.scale_dirty(scale)
        queue.finish()
        if i == 0:
            writer.write_fits_grid('grid', 'grid', not args.host, grid_data, image_p, channel)
            writer.write_fits_image('dirty', 'dirty image', dataset, dirty, image_p,
                                    channel, restoring_beam)

        # Deconvolution
        noise = imager.noise_est()
        imager.clean_reset()
        peak_value = imager.clean_cycle(psf_patch)
        peak_power = clean.metric_to_power(clean_p.mode, peak_value)
        noise_threshold = noise * clean_p.threshold
        mgain_threshold = (1.0 - clean_p.major_gain) * peak_power
        logger.info('Threshold from noise estimate: %g', noise_threshold)
        logger.info('Threshold from mgain:          %g', mgain_threshold)
        threshold = max(noise_threshold, mgain_threshold)
        if peak_power <= threshold:
            logger.info('Threshold reached, terminating')
            break
        logger.info('CLEANing to threshold:         %g', threshold)
        threshold_metric = clean.power_to_metric(clean_p.mode, threshold)
        bar = progress.make_progressbar('CLEAN', max=clean_p.minor - 1)
        with progress.finishing(bar):
            for j in bar.iter(range(clean_p.minor - 1)):
                value = imager.clean_cycle(psf_patch, threshold_metric)
                if value is None:
                    break
        queue.finish()

    writer.write_fits_image('model', 'model', dataset, model, image_p, channel)
    writer.write_fits_image('residuals', 'residuals', dataset, dirty, image_p,
                            channel, restoring_beam)

    # Try to free up memory for the beam convolution
    del grid_data
    del psf
    del imager

    # Convolve with restoring beam, and add residuals back in
    if args.host:
        beam.convolve_beam(model, restoring_beam, model)
    else:
        restore = beam.ConvolveBeamTemplate(queue, model.shape[1:], model.dtype).instantiate()
        restore.beam = restoring_beam
        restore.ensure_all_bound()
        restore_image = restore.buffer('image')
        for pol in range(model.shape[0]):
            # TODO: eliminate these copies, and work directly on model
            model.copy_region(queue, restore_image, np.s_[pol], ())
            restore()
            restore_image.copy_region(queue, model, (), np.s_[pol])
        queue.finish()

    model += dirty
    writer.write_fits_image('clean', 'clean image', dataset, model, image_p,
                            channel, restoring_beam)


def run(args, context, queue, dataset, writer):
    # PyCUDA leaks resources that are freed when the corresponding context is
    # not active. We make it active for the rest of the execution to avoid
    # this.
    with context:
        # Determine parameters
        input_polarizations = dataset.polarizations()
        if dataset.has_feed_angles():
            polarization_matrices = \
                polarization.polarization_matrices(args.stokes, input_polarizations)
        else:
            polarization_matrices = (
                polarization.polarization_matrix(args.stokes, input_polarizations), None)
        array_p = dataset.array_parameters()
        if args.stop_channel is None:
            args.stop_channel = dataset.num_channels()
        if not (0 <= args.start_channel < args.stop_channel <= dataset.num_channels()):
            raise ValueError('Channels are out of range')
        weight_p = parameters.WeightParameters(
            weight.WeightType[args.weight_type.upper()], args.robustness)

        if args.stop_channel - args.start_channel > 1:
            ChannelParameters(
                args, dataset, args.start_channel, array_p).log_parameters(' [first channel]')
            ChannelParameters(
                args, dataset, args.stop_channel - 1, array_p).log_parameters(' [last channel]')
        else:
            ChannelParameters(args, dataset, args.start_channel, array_p).log_parameters()
        log_parameters("Weight parameters", weight_p)

        if args.subtract == 'auto':
            subtract_model = dataset.sky_model()
        elif args.subtract is not None:
            subtract_model = sky_model.open_sky_model(args.subtract)
        else:
            subtract_model = None

        for start_channel in range(args.start_channel, args.stop_channel, args.channel_batch):
            stop_channel = min(args.stop_channel, start_channel + args.channel_batch)
            channels = range(start_channel, stop_channel)
            params = [ChannelParameters(args, dataset, channel, array_p)
                      for channel in channels]
            # Preprocess visibilities
            image_ps = [channel_p.image_p for channel_p in params]
            grid_ps = [channel_p.grid_p for channel_p in params]
            collector = preprocess_visibilities(dataset, args, start_channel, stop_channel,
                                                image_ps, grid_ps, polarization_matrices)
            reader = collector.reader()

            # Do the work
            for channel_p in params:
                process_channel(dataset, args, start_channel, context, queue,
                                reader, writer, channel_p, array_p, weight_p, subtract_model)



#!/usr/bin/env python
from setuptools import setup, find_packages, Extension
import glob
import os


class MissingPkgconfig:
    """Raise an exception only when trying to convert it to a string.

    This allows commands like ``setup.py clean`` to work even when pkgconfig
    is not present, but makes it fail when trying to build the extension.
    """
    def __str__(self):
        raise RuntimeError(
            'The pkgconfig module was not found. Try upgrading pip to the latest '
            'version and using it to install.')


try:
    import pkgconfig
    eigen3 = pkgconfig.parse('eigen3')
except ImportError:
    eigen3 = {'include_dirs': {MissingPkgconfig()}}

tests_require = ['nose', 'scipy', 'fakeredis']

root_dir = os.path.dirname(__file__)
pybind11_dir = os.path.join(root_dir, '3rdparty', 'pybind11', 'include', 'pybind11')
if not os.path.exists(pybind11_dir):
    raise RuntimeError(
        'pybind11 directory not found in source tree. If this is a git checkout, you '
        'can probably fix it by running "git submodule update --init --recursive".')

with open(os.path.join(root_dir, 'README.rst')) as f:
    long_description = f.read()

extensions = [
    Extension(
        '_preprocess',
        sources=['katsdpimager/preprocess.cpp'],
        language='c++',
        include_dirs=['3rdparty/pybind11/include'] + list(eigen3.get('include_dirs', [])),
        depends=glob.glob('katsdpimager/*.h'),
        extra_compile_args=['-std=' + os.environ.get('KATSDPIMAGER_STD_CXX', 'c++1y'),
                            '-g0', '-fvisibility=hidden'],
        libraries=list(eigen3.get('libraries', []))
    )
]

setup(
    name="katsdpimager",
    description="GPU-accelerated radio-astronomy spectral line imager",
    long_description=long_description,
    author="MeerKAT SDP Team",
    author_email="sdpdev+katsdpimager@ska.ac.za",
    url="https://github.com/ska-sa/katsdpimager/",
    packages=find_packages(),
    package_data={'': ['imager_kernels/*.mako', 'imager_kernels/*/*.mako']},
    scripts=["scripts/imager.py", "scripts/imager-mkat-pipeline.py", "scripts/fits-video.py"],
    ext_package='katsdpimager',
    ext_modules=extensions,
    python_requires='>=3.6',
    install_requires=[
        'numpy>=1.10.0', 'katsdpsigproc', 'katpoint', 'astropy>=1.3', 'progress',
        'pycuda', 'scikit-cuda', 'h5py', 'ansicolors'
    ],
    tests_require=tests_require,
    extras_require={
        'test': tests_require,
        'doc': ['sphinx>=1.3.0', 'sphinxcontrib-tikz', 'sphinx-rtd-theme'] + tests_require,
        'cpu': ['numba'],
        'report': ['aplpy', 'matplotlib', 'mako'],
        'ms': ['python-casacore'],
        'katdal': ['katdal', 'scipy>=0.17'],
        'benchmark': ['katpoint'],
        'pipeline': ['katsdpservices', 'matplotlib']
    },
    use_katversion=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering :: Astronomy",
        "Environment :: Console"
    ]
)



"""Tests for :mod:`katsdpimager.grid`."""

from unittest import mock

import numpy as np
import astropy.units as units
import katsdpsigproc.accel as accel
from katsdpsigproc.test.test_accel import device_test, force_autotune

from .. import parameters, polarization, grid
from .utils import RandomState


def _middle(array, shape):
    """Returns a view of the central part of `array` with size `shape`."""
    index = []
    for a, s in zip(array.shape, shape):
        assert a >= s
        assert (a - s) % 2 == 0
        pad = (a - s) // 2
        index.append(np.s_[pad : a - pad])
    return array[tuple(index)]


class BaseTest:
    def setup(self):
        pixels = 256
        grid_cover = 180
        w_planes = 32
        oversample = 8
        n_vis = 1000
        kernel_width = 28
        assert grid_cover + kernel_width < pixels
        self.image_parameters = parameters.ImageParameters(
            q_fov=1.0,
            image_oversample=None,
            frequency=0.01 * units.m,
            array=None,
            polarizations=polarization.STOKES_IQUV,
            dtype=np.float64,
            pixel_size=0.0001,
            pixels=pixels)
        # Make a single-precision version for cases that don't need double
        self.image_parameters_sp = parameters.ImageParameters(
            q_fov=1.0,
            image_oversample=None,
            frequency=self.image_parameters.wavelength,
            array=None,
            polarizations=self.image_parameters.polarizations,
            dtype=np.float32,
            pixel_size=self.image_parameters.pixel_size,
            pixels=self.image_parameters.pixels)
        self.grid_parameters = parameters.GridParameters(
            antialias_width=7.0,
            oversample=oversample,
            image_oversample=4,
            w_slices=1,
            w_planes=w_planes,
            max_w=5 * units.m,
            kernel_width=kernel_width)
        self.array_parameters = mock.Mock()
        self.array_parameters.longest_baseline = self.image_parameters.cell_size * (grid_cover // 2)
        # Create a track in which movement happens in various subsets of the
        # axes, using a random walk
        rs = np.random.RandomState(seed=1)
        self.uv = np.empty((n_vis, 2), dtype=np.int16)
        self.sub_uv = np.empty((n_vis, 2), dtype=np.int16)
        self.w_plane = np.empty((n_vis,), dtype=np.int16)
        for i in range(0, n_vis):
            if i % 73 == 0:
                # Add an occasional total jump
                self.uv[i, :] = rs.randint(0, grid_cover, (2,))
                self.sub_uv[i, :] = rs.randint(0, oversample, (2,))
                self.w_plane[i] = rs.randint(0, w_planes)
            else:
                for j in range(2):
                    self.uv[i, j] = \
                        (self.uv[i - 1, j] + rs.random_integers(-1, 1)) % grid_cover
                    self.sub_uv[i, j] = \
                        (self.sub_uv[i - 1, j] + rs.random_integers(-1, 1)) % oversample
                self.w_plane[i] = (self.w_plane[i - 1] + rs.random_integers(-1, 1)) % w_planes
        self.uv -= grid_cover // 2
        self.convolve_kernel = grid.ConvolutionKernel(
            self.image_parameters, self.grid_parameters)
        self.weights_grid = rs.uniform(size=(4, grid_cover, grid_cover)).astype(np.float32)

    def do_grid(self, callback):
        max_vis = 1280
        n_vis = len(self.w_plane)
        rs = RandomState(seed=2)
        vis = rs.complex_uniform(-1, 1, size=(n_vis, 4)).astype(np.complex128)
        actual = callback(max_vis, vis)
        expected = np.zeros_like(actual)
        pixels = actual.shape[-1]
        uv_bias = (self.convolve_kernel.data.shape[-1] - 1) // 2 - pixels // 2
        for i in range(n_vis):
            kernel = np.outer(self.convolve_kernel.data[self.w_plane[i], self.sub_uv[i, 1], :],
                              self.convolve_kernel.data[self.w_plane[i], self.sub_uv[i, 0], :])
            kernel = np.conj(kernel, kernel)
            u = self.uv[i, 0] - uv_bias
            v = self.uv[i, 1] - uv_bias
            weights_u = self.uv[i, 0] + self.weights_grid.shape[2] // 2
            weights_v = self.uv[i, 1] + self.weights_grid.shape[1] // 2
            for j in range(4):
                footprint = expected[j, v : v + kernel.shape[0], u : u + kernel.shape[1]]
                weight = self.weights_grid[j, weights_v, weights_u]
                footprint[:] += vis[i, j].astype(np.complex128) * weight * kernel
        np.testing.assert_allclose(expected, actual, 1e-5, 1e-8)

    def do_degrid(self, callback):
        max_vis = 1280
        n_vis = len(self.w_plane)
        pixels = self.image_parameters.pixels
        shape = (4, pixels, pixels)
        rs = RandomState(seed=2)
        grid_data = rs.complex_uniform(-1, 1, size=shape).astype(np.complex128)
        vis = rs.complex_uniform(-1, 1, size=(n_vis, 4)).astype(np.complex128)
        weights = rs.uniform(0.5, 1.5, size=(n_vis, 4)).astype(np.float64)
        expected = np.zeros_like(vis.copy())
        uv_bias = (self.convolve_kernel.data.shape[-1] - 1) // 2 - pixels // 2
        for i in range(n_vis):
            kernel = np.outer(self.convolve_kernel.data[self.w_plane[i], self.sub_uv[i, 1], :],
                              self.convolve_kernel.data[self.w_plane[i], self.sub_uv[i, 0], :])
            u = self.uv[i, 0] - uv_bias
            v = self.uv[i, 1] - uv_bias
            for j in range(4):
                footprint = grid_data[j, v : v + kernel.shape[0], u : u + kernel.shape[1]]
                expected[i, j] = (vis[i, j]
                                  - weights[i, j] * np.dot(kernel.ravel(), footprint.ravel()))
        residuals = callback(max_vis, grid_data, weights, vis)
        np.testing.assert_allclose(expected, residuals, 1e-5)


class TestGridder(BaseTest):
    """Tests for :class:`grid.Gridder`"""
    @device_test
    def test(self, context, command_queue):
        def callback(max_vis, vis):
            template = grid.GridderTemplate(context, self.image_parameters, self.grid_parameters)
            fn = template.instantiate(command_queue, self.array_parameters, max_vis,
                                      allocator=accel.SVMAllocator(context))
            fn.ensure_all_bound()
            grid_data = fn.buffer('grid')
            grid_data.fill(0)
            weights_grid = fn.buffer('weights_grid')
            weights_grid_host = weights_grid.empty_like()
            _middle(weights_grid_host, self.weights_grid.shape)[:] = self.weights_grid
            weights_grid.set_async(command_queue, weights_grid_host)
            fn.num_vis = len(self.uv)
            fn.set_coordinates(self.uv, self.sub_uv, self.w_plane)
            fn.set_vis(vis)
            fn()
            command_queue.finish()
            return grid_data
        self.do_grid(callback)

    @device_test
    @force_autotune
    def test_autotune(self, context, queue):
        """Check that the autotuner runs successfully"""
        grid.GridderTemplate(context, self.image_parameters_sp, self.grid_parameters)


class TestGridderHost(BaseTest):
    """Tests for :Class:`grid.GridderHost`"""
    def test(self):
        def callback(max_vis, vis):
            gridder = grid.GridderHost(self.image_parameters, self.grid_parameters)
            gridder.clear()
            _middle(gridder.weights_grid, self.weights_grid.shape)[:] = self.weights_grid
            gridder.num_vis = len(self.uv)
            gridder.set_coordinates(self.uv, self.sub_uv, self.w_plane)
            gridder.set_vis(vis)
            gridder()
            return gridder.values
        self.do_grid(callback)


class TestDegridder(BaseTest):
    """Tests for :class:`grid.Degridder`"""
    @device_test
    def test(self, context, command_queue):
        def callback(max_vis, grid_data, weights, vis):
            n_vis = len(self.w_plane)
            template = grid.DegridderTemplate(context, self.image_parameters, self.grid_parameters)
            fn = template.instantiate(command_queue, self.array_parameters, max_vis,
                                      allocator=accel.SVMAllocator(context))
            fn.ensure_all_bound()
            grid_buffer = fn.buffer('grid')
            grid_buffer.set(command_queue, _middle(grid_data, grid_buffer.shape))
            fn.num_vis = n_vis
            fn.set_coordinates(self.uv, self.sub_uv, self.w_plane)
            fn.set_weights(weights)
            fn.set_vis(vis)
            fn()
            command_queue.finish()
            return fn.buffer('vis')[:n_vis]
        self.do_degrid(callback)

    @device_test
    @force_autotune
    def test_autotune(self, context, queue):
        """Check that the autotuner runs successfully"""
        grid.DegridderTemplate(context, self.image_parameters_sp, self.grid_parameters)


class TestDegridderHost(BaseTest):
    """Tests for :class:`grid.DegridderHost`"""
    def test(self):
        def callback(max_vis, grid_data, weights, vis):
            n_vis = len(self.w_plane)
            degridder = grid.DegridderHost(self.image_parameters, self.grid_parameters)
            degridder.values[:] = grid_data
            degridder.num_vis = n_vis
            degridder.set_coordinates(self.uv, self.sub_uv, self.w_plane)
            degridder.set_weights(weights)
            degridder.set_vis(vis)
            degridder()
            return vis
        self.do_degrid(callback)



"""Tests for :py:mod:`katsdpimager.clean`"""

import numpy as np
import scipy.signal
from katsdpsigproc.test.test_accel import device_test
from nose.tools import assert_equal, assert_greater_equal

from .. import clean


class _TestPsfPatchBase:
    """Tests for :class:`~katsdpimager.clean.PsfPatch` and :class:`~katsdpimager.clean.psf_patch_host."""   # noqa: E501
    def test_peak_only(self):
        assert_equal((4, 1, 1), self._test())

    def test_low_corner(self):
        self.psf_host[0, 0, 0] = 0.1
        assert_equal((4, 206, 304), self._test())

    def test_high_corner(self):
        self.psf_host[3, 205, 303] = -0.2
        assert_equal((4, 205, 303), self._test())

    def test_1d(self):
        target = self.psf_host[1, 0, :152]
        target[:] = np.arange(152)
        threshold = 50.5
        box = self._test(threshold=threshold)
        hw = box[2] // 2
        assert_equal(0, sum(target[:-hw] >= threshold))
        assert_greater_equal(target[-hw], threshold)

    def test_limit(self):
        self.psf_host[0, 0, 0] = 0.4
        self.psf_host[3, 205, 303] = 0.3
        self.psf_host[1, 110, 150] = 0.2
        assert_equal((4, 15, 5), self._test(limit=50))


class TestPsfPatch(_TestPsfPatchBase):
    @device_test
    def setup(self, context, command_queue):
        self.template = clean.PsfPatchTemplate(context, np.float32, 4)
        self.fn = self.template.instantiate(command_queue, (4, 206, 304))
        self.fn.ensure_all_bound()
        self.psf = self.fn.buffer('psf')
        self.psf_host = self.psf.empty_like()
        self.psf_host.fill(0.0)
        self.psf_host[:, 103, 152] = 1.0    # Set central peak

    def _test(self, threshold=0.01, limit=None):
        """Run the function using psf_host, returning the resulting patch size."""
        self.psf.set(self.fn.command_queue, self.psf_host)
        return self.fn(threshold, limit)


class TestPsfPatchHost(_TestPsfPatchBase):
    def setup(self):
        self.psf_host = np.zeros((4, 206, 304), np.float32)

    def _test(self, threshold=0.01, limit=None):
        return clean.psf_patch_host(self.psf_host, threshold, limit)


class TestClean:
    @classmethod
    def _make_psf(cls, size):
        """Creates a dummy PSF as a Gaussian plus random noise"""
        rs = np.random.RandomState(seed=1)
        gaussian1d = scipy.signal.gaussian(size, size / 8)
        psf = np.outer(gaussian1d, gaussian1d)
        psf[:] += rs.standard_normal(psf.shape)
        return psf.astype(np.float32)

    @classmethod
    def _zero_buffer(cls, command_queue, buf):
        buf.set(command_queue, np.zeros(buf.shape, buf.dtype))

    @device_test
    def test_update_tiles(self, context, command_queue):
        image_shape = (4, 567, 456)
        border = 65
        rs = np.random.RandomState(seed=1)
        template = clean._UpdateTilesTemplate(context, np.float32, 4, clean.CLEAN_I)
        fn = template.instantiate(command_queue, image_shape, border)
        fn.ensure_all_bound()
        # TODO: this could lead to ties, which will fail the test
        dirty = rs.standard_normal(image_shape).astype(np.float32)
        fn.buffer('dirty').set(command_queue, dirty)
        self._zero_buffer(command_queue, fn.buffer('tile_max'))
        self._zero_buffer(command_queue, fn.buffer('tile_pos'))
        fn(135, 161, 385, 450)
        tile_max = fn.buffer('tile_max').get(command_queue)
        tile_pos = fn.buffer('tile_pos').get(command_queue)

        num_tiles_y, num_tiles_x = tile_max.shape
        assert_equal(14, num_tiles_y)
        assert_equal(11, num_tiles_x)
        for y in range(num_tiles_y):
            for x in range(num_tiles_x):
                # Make sure updates don't happen where they are not meant to
                if x < 2 or x >= 10 or y < 3 or y >= 13:
                    assert_equal(0.0, tile_max[y, x])
                    assert_equal(0, tile_pos[y, x, 0])
                    assert_equal(0, tile_pos[y, x, 1])
                else:
                    y0 = y * template.tiley + border
                    x0 = x * template.tilex + border
                    y1 = min(y0 + template.tiley, dirty.shape[1] - border)
                    x1 = min(x0 + template.tilex, dirty.shape[2] - border)
                    tile = np.abs(dirty[0, y0:y1, x0:x1])
                    pos = np.unravel_index(np.argmax(tile), tile.shape)
                    assert_equal(tile[pos], tile_max[y, x])
                    assert_equal(pos[0] + y0, tile_pos[y, x, 0])
                    assert_equal(pos[1] + x0, tile_pos[y, x, 1])

    @device_test
    def test_find_peak(self, context, command_queue):
        image_shape = (4, 256, 256)
        tile_shape = (72, 67)
        template = clean._FindPeakTemplate(context, np.float32, 4)
        fn = template.instantiate(command_queue, image_shape, tile_shape)
        fn.ensure_all_bound()
        # TODO: this could lead to ties, which will fail the test
        rs = np.random.RandomState(seed=1)
        dirty = rs.uniform(1.0, 2.0, image_shape).astype(np.float32)
        tile_max = rs.uniform(1.0, 2.0, tile_shape).astype(np.float32)
        # Set tile_pos so that each pos indexes the corresponding position in
        # the dirty image, to make testing easier.
        tile_pos = np.array(
            [[[y, x] for x in range(tile_shape[1])] for y in range(tile_shape[0])], np.int32)
        fn.buffer('tile_max').set(command_queue, tile_max)
        fn.buffer('tile_pos').set(command_queue, tile_pos)
        fn.buffer('dirty').set(command_queue, dirty)
        fn()
        peak_value = fn.buffer('peak_value').get(command_queue)
        peak_pos = fn.buffer('peak_pos').get(command_queue)
        peak_pixel = fn.buffer('peak_pixel').get(command_queue)
        best = np.unravel_index(np.argmax(tile_max), tile_max.shape)
        assert_equal(tile_max[best], peak_value[0])
        np.testing.assert_array_equal(tile_pos[best], peak_pos)
        np.testing.assert_array_equal(dirty[:, peak_pos[0], peak_pos[1]], peak_pixel)

    @device_test
    def test_subtract_psf(self, context, command_queue):
        loop_gain = 0.25
        image_shape = (4, 200, 344)
        psf_patch = (4, 72, 130)
        pos = (170, 59)    # chosen so that the PSF will be clipped to the image
        rs = np.random.RandomState(seed=1)
        dirty = rs.standard_normal(image_shape).astype(np.float32)
        psf = rs.standard_normal(psf_patch).astype(np.float32)
        expected = dirty.copy()
        peak_pixel = dirty[:, pos[0], pos[1]]
        expected[:, 134:200, 0:124] -= \
            loop_gain * peak_pixel[:, np.newaxis, np.newaxis] * psf[:, :66, 6:]
        psf_full = np.ones(image_shape, np.float32)
        psf_full[:, 64:136, 107:237] = psf

        template = clean._SubtractPsfTemplate(context, np.float32, 4)
        fn = template.instantiate(command_queue, loop_gain, image_shape, image_shape)
        fn.ensure_all_bound()
        fn.buffer('dirty').set(command_queue, dirty)
        fn.buffer('psf').set(command_queue, psf_full)
        fn.buffer('peak_pixel').set(command_queue, peak_pixel)
        self._zero_buffer(command_queue, fn.buffer('model'))
        fn(pos, psf_patch)
        dirty = fn.buffer('dirty').get(command_queue)
        model = fn.buffer('model').get(command_queue)
        np.testing.assert_allclose(expected, dirty, atol=1e-4)
        np.testing.assert_allclose(loop_gain * peak_pixel, model[:, pos[0], pos[1]])

    def _test_noise(self, context, command_queue, std, rtol):
        rs = np.random.RandomState(seed=1)
        image_shape = (4, 400, 544)
        border = 45
        # Start with a standard normal distribution
        dirty = rs.standard_normal(image_shape).astype(np.float32)
        # Scale it up only inside the border, to ensure that the border value
        # is being respected.
        dirty[:, border:-border, border:-border] *= std
        # Add some big values to ensure robustness
        dirty.flat[rs.choice(dirty.size, 1000, replace=False)] += 1e6

        # Test with CLEAN_I
        template = clean.NoiseEstTemplate(context, np.float32, 4, clean.CLEAN_I)
        fn = template.instantiate(command_queue, image_shape, border)
        fn.ensure_all_bound()
        fn.buffer('dirty').set(command_queue, dirty)
        estimated = fn()
        np.testing.assert_allclose(estimated, std, rtol=rtol)

        # Test with CLEAN_SUMSQ
        template = clean.NoiseEstTemplate(context, np.float32, 4, clean.CLEAN_SUMSQ)
        fn = template.instantiate(command_queue, image_shape, border)
        fn.ensure_all_bound()
        fn.buffer('dirty').set(command_queue, dirty)
        estimated = fn()
        # Magic number is the ratio of the medians of the chi distribution with
        # 4 and 1 degrees of freedom:
        # scipy.stats.chi(4).median() / scipy.stats.chi(1).median()
        np.testing.assert_allclose(estimated, std * 2.716317430527251, rtol=rtol)

    @device_test
    def test_noise(self, context, command_queue):
        self._test_noise(context, command_queue, 3.2, 1e-2)

    @device_test
    def test_noise_zero(self, context, command_queue):
        self._test_noise(context, command_queue, 0.0, 0.0)



"""Render images using matplotlib."""

import numpy as np
from astropy import units
from astropy.wcs import WCS
import astropy.io.fits as fits
import matplotlib
matplotlib.use('Agg')     # noqa: E402
import matplotlib.axes
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from mpl_toolkits.axes_grid1 import make_axes_locatable

import katsdpsigproc.zscale as zscale


DEFAULT_DPI = 96


def _prepare_axes(wcs, width, height, image_width, image_height, dpi):
    fig = plt.Figure(figsize=(width / dpi, height / dpi), dpi=dpi)
    ax = fig.add_subplot(projection=wcs, slices=('x', 'y', 0, 0))
    ax.set_xlabel('Right Ascension')
    ax.set_ylabel('Declination')
    ax.set_xlim(-0.5, image_width - 0.5)
    ax.set_ylim(-0.5, image_height - 0.5)
    return fig, ax


def _plot(hdus, caption, ax, extent, vmin, vmax):
    bunit = hdus[0].header['BUNIT']
    if bunit == 'JY/BEAM':
        # This was used in older versions of katsdpimager, but is not FITS-standard
        unit = units.Jy / units.beam
    else:
        unit = units.Unit(bunit)
    data = hdus[0].data[0, 0, :, :] << unit
    vmin <<= unit
    vmax <<= unit
    # If the flux is low, use µJy/beam or mJy/beam to keep scale sane
    if vmax < 100 * (units.uJy / units.beam):
        data = data.to(units.uJy / units.beam)
    elif vmax < 100 * (units.mJy / units.beam):
        data = data.to(units.mJy / units.beam)
    vmin = vmin.to(data.unit)
    vmax = vmax.to(data.unit)
    if not ax.images:
        im = ax.imshow(data.value, origin='lower', cmap='afmhot', aspect='equal',
                       vmin=vmin.value, vmax=vmax.value, extent=extent)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes('right', pad='3%', size='5%', axes_class=matplotlib.axes.Axes)
        cbar = ax.get_figure().colorbar(im, cax=cax, orientation='vertical')
        # simply using data.unit.to_string('unicode') draws an ASCII-art
        # fraction, which doesn't end up looking very good. But we want
        # unicode to properly render µJy (rather than uJy).
        unit_label = (data.unit * units.beam).to_string('unicode') + ' / beam'
        cbar.set_label(unit_label)
    else:
        im = ax.images[0]
        im.set_data(data)
        im.set_extent(extent)

    # Convert an arbitrary pixel to world coordinates to get frequency
    wcs = WCS(hdus[0])
    pix = np.array([[0, 0, 0, 0]])
    world = wcs.all_pix2world(pix, 0)
    freq = world[0][3] * units.Unit(hdus[0].header['CUNIT4'])
    freq = freq.to(units.MHz)
    if caption:
        ax.set_title(f'{caption} ({freq:.3f})')
    else:
        ax.set_title(f'{freq:.3f}')


def write_image(input_file, output_file, width=1024, height=768, dpi=DEFAULT_DPI, caption=None):
    """Write an image showing a single FITS file.

    This code is only designed to work with FITS files written by
    :func:`katsdpimager.io.write_fits_image` (it makes assumptions about axis
    order, units etc).

    Parameters
    ----------
    input_file : str
        Source FITS file
    output_file : str
        Output image file, including extension
    width, height : int
        Dimensions of the output image
    caption : Optional[str]
        Optional caption to include in the image
    """
    with fits.open(input_file) as hdus:
        data = hdus[0].data[0, 0]    # Stokes I, single channel
        vmin, vmax = zscale.zscale(zscale.sample_image(data))
        image_height, image_width = data.shape
        fig, ax = _prepare_axes(WCS(hdus[0]), width, height, image_width, image_height, dpi)
        _plot(hdus, caption, ax, None, vmin, vmax)
        fig.savefig(output_file)


def write_movie(files, output_file, width=1024, height=768, dpi=DEFAULT_DPI, fps=5.0):
    """Write a video with an animation of a set of FITS files.

    This code is only designed to work with FITS files written by
    :func:`katsdpimager.io.write_fits_image` (it makes assumptions about axis
    order, units etc).

    The field of view is determined by the last image. With increasing
    frequency, this is usually the one with the smallest field of view,
    ensuring that the frame is filled in all the images (with pixels
    clipped from the channels with a wider field of view).

    Note that matplotlib currently uses ffmpeg (the program) to write a video,
    so it needs to be installed.

    Parameters
    ----------
    files : Sequence[Tuple[Optional[str], str]]
        Pairs of caption and filename.
    output_file : str
        Output filename, including extension.
    width, height : int
        Nominal dimensions of the video. Due to limitations in matplotlib
        it might not be exact.
    fps : float
        Frames per second in the written video.
    """
    # Load the last image to get its WCS
    with fits.open(files[-1][1]) as hdus:
        common_wcs = WCS(hdus[0])
        image_width = hdus[0].header['NAXIS1']
        image_height = hdus[0].header['NAXIS2']
    # Sample all the images to choose data bounds
    samples = []
    n_samples = 1000000 // len(files) + 1
    for caption, filename in files:
        with fits.open(filename, memmap=True) as hdus:
            s = zscale.sample_image(hdus[0].data[0, 0], max_samples=n_samples, random_offsets=True)
            samples.append(s)
    samples = np.concatenate(samples)
    vmin, vmax = zscale.zscale(samples)
    fig, ax = _prepare_axes(common_wcs, width, height, image_width, image_height, dpi)

    def render_channel(caption_filename):
        caption, filename = caption_filename
        with fits.open(filename) as hdus:
            wcs = WCS(hdus[0])
            data = hdus[0].data[0, 0, :, :]
            # Convert corners of the image to world coordinates
            corners_pix = np.array([
                [-0.5, -0.5, 0, 0],
                [data.shape[1] - 0.5, data.shape[0] - 0.5, 0, 0]
            ])
            corners_world = wcs.all_pix2world(corners_pix, 0)
            # Convert back to pixel coordinates for the plotting WCS
            corners_data = common_wcs.all_world2pix(corners_world, 0)
            extent = [corners_data[0, 0], corners_data[1, 0],
                      corners_data[0, 1], corners_data[1, 1]]
            _plot(hdus, caption, ax, extent, vmin, vmax)

    ani = animation.FuncAnimation(fig, render_channel, files, cache_frame_data=False)
    ani.save(output_file, fps=fps)



"""Tests for :py:mod:`katsdpimager.preprocess`."""

import tempfile
import logging
import os
from contextlib import closing

import astropy.units as units
import numpy as np
from nose.tools import assert_equal, assert_true, assert_false

import katsdpimager.parameters as parameters
import katsdpimager.polarization as polarization
import katsdpimager.preprocess as preprocess


def _empty_recarray(dtype):
    return np.rec.array(None, dtype=dtype, shape=(0,))


class BaseTestVisibilityCollector:
    def setup(self):
        self.image_parameters = []
        self.grid_parameters = []
        for wavelength in np.array([0.25, 0.125]) * units.m:
            self.image_parameters.append(parameters.ImageParameters(
                q_fov=1.0,
                image_oversample=5.0,
                frequency=wavelength, array=None,
                polarizations=polarization.STOKES_IQUV,
                dtype=np.float32,
                pixel_size=1.0/(4096.0*wavelength.value),
                pixels=2048))
            self.grid_parameters.append(parameters.GridParameters(
                antialias_width=7.0,
                oversample=8,
                image_oversample=4.0,
                w_slices=1,
                w_planes=128,
                max_w=400 * units.m,
                kernel_width=64))

    def check(self, collector, expected):
        reader = collector.reader()
        assert_equal(reader.num_channels, collector.num_channels)
        for channel, channel_data in enumerate(expected):
            for w_slice, slice_data in enumerate(channel_data):
                assert_equal(len(slice_data), reader.len(channel, w_slice))
                for block_size in [None, 1, 2, 100]:
                    pieces = []
                    for piece in reader.iter_slice(channel, w_slice, block_size):
                        pieces.append(piece.copy())
                    if pieces:
                        actual = np.rec.array(np.hstack(pieces))
                    else:
                        actual = np.rec.recarray(0, collector.store_dtype)
                    # np.testing functions don't seem to handle structured arrays
                    # well, so test one field at a time.
                    np.testing.assert_array_equal(actual.uv, slice_data.uv)
                    np.testing.assert_array_equal(actual.sub_uv, slice_data.sub_uv)
                    np.testing.assert_allclose(actual.weights, slice_data.weights)
                    np.testing.assert_allclose(actual.vis, slice_data.vis, rtol=1e-5)
                    np.testing.assert_array_equal(actual.w_plane, slice_data.w_plane)

    def test_empty(self):
        with closing(self.factory(self.image_parameters, self.grid_parameters, 2)) as collector:
            pass
        self.check(collector, [[np.rec.recarray(0, collector.store_dtype)] for channel in range(2)])

    def _test_impl(self, use_feed_angles):
        uvw = np.array([
            [12.1, 2.3, 4.7],
            [-3.4, 7.6, 2.5],
            [-5.2, -10.6, 7.2],
            [12.102, 2.299, 4.6],   # Should merge with the first visibility in first channel
            [-1.0, 2.0, 3.0]
        ], dtype=np.float32) * units.m
        weights = np.array([
            [
                [1.3, 0.6, 1.2, 0.1],
                [1.0, 1.0, 1.0, 1.0],
                [0.5, 0.6, 0.7, 0.8],
                [1.1, 1.2, 1.3, 1.4],
                [1.0, 0.0, 1.0, 1.0]    # Has a zero weight, so should be skipped
            ], [
                [0.2, 2.4, 1.2, 2.6],   # Second channel is double and reverse of first
                [2.0, 2.0, 2.0, 2.0],
                [1.6, 1.4, 1.2, 1.0],
                [2.8, 2.6, 2.4, 2.2],
                [2.0, 2.0, 0.0, 2.0]
            ]], dtype=np.float32)
        baselines = np.array([
            0,
            -1,    # Auto-correlation: should be removed
            1,
            0,
            0
        ], dtype=np.int16)
        vis = np.array([
            [
                [0.5 - 2.3j, 0.1 + 4.2j, 0.0 - 3j, 1.5 + 0j],
                [0.5 - 2.3j, 0.1 + 4.2j, 0.0 - 3j, 1.5 + 0j],
                [1.5 + 1.3j, 1.1 + 2.7j, 1.0 - 2j, 2.5 + 1j],
                [1.2 + 3.4j, 5.6 + 7.8j, 9.0 + 1.2j, 3.4 + 5.6j],
                [10.0, 10.0, 10.0, 10.0]
            ], [
                [3.0 + 0j, 0.0 - 6j, 0.2 + 8.4j, 1.0 - 4.6j],
                [3.0 + 0j, 0.0 - 6j, 0.2 + 8.4j, 1.0 - 4.6j],
                [3.0 + 2j, 2.0 - 4j, 2.2 + 5.4j, 3.0 + 2.6j],
                [6.8 + 11.2j, 18.0 + 2.4j, 11.2 + 15.6j, 2.4 + 6.8j],
                [20.0, 20.0, 20.0, 20.0]
            ]], dtype=np.complex64)
        if use_feed_angles:
            # TODO: use non-trivial feed angles and matrices
            feed_angle1 = feed_angle2 = np.zeros(5, np.float32)
            mueller_stokes = mueller_circular = np.matrix(np.identity(4, np.complex64))
        else:
            feed_angle1 = feed_angle2 = mueller_circular = None
            mueller_stokes = np.matrix(np.identity(4, np.complex64))
        with closing(self.factory(self.image_parameters, self.grid_parameters, 64)) as collector:
            collector.add(uvw, weights, baselines, vis,
                          feed_angle1, feed_angle2,
                          mueller_stokes, mueller_circular)
        self.check(collector, [
            [np.rec.fromarrays([
                [[96, 18], [-42, -85]],
                [[6, 3], [3, 1]],
                [[2.4, 1.8, 2.5, 1.5], [0.5, 0.6, 0.7, 0.8]],
                [[1.97 + 0.75j, 6.78 + 11.88j, 11.7 - 2.04j, 4.91 + 7.84j],
                 [0.75 + 0.65j, 0.66 + 1.62j, 0.7 - 1.4j, 2.0 + 0.8j]],
                [64, 65]], dtype=collector.store_dtype)],
            [np.rec.fromarrays([
                [[387, 73], [387, 73], [-167, -340]],
                [[1, 4], [2, 4], [4, 6]],
                [[0.2, 2.4, 1.2, 2.6], [2.8, 2.6, 2.4, 2.2], [1.6, 1.4, 1.2, 1.0]],
                [[0.6 + 0.0j, 0.0 - 14.4j, 0.24 + 10.08j, 2.6 - 11.96j],
                 [19.04 + 31.36j, 46.8 + 6.24j, 26.88 + 37.44j, 5.28 + 14.96j],
                 [4.8 + 3.2j, 2.8 - 5.6j, 2.64 + 6.48j, 3.0 + 2.6j]],
                [64, 64, 65]], dtype=collector.store_dtype)]
        ])

    def test_simple(self):
        self._test_impl(False)

    def test_feed_angles(self):
        self._test_impl(True)


def test_is_prime():
    assert_true(preprocess._is_prime(2))
    assert_true(preprocess._is_prime(3))
    assert_true(preprocess._is_prime(11))
    assert_true(preprocess._is_prime(10007))
    assert_false(preprocess._is_prime(4))
    assert_false(preprocess._is_prime(6))
    assert_false(preprocess._is_prime(18))
    assert_false(preprocess._is_prime(21))


class TestVisibilityCollectorMem(BaseTestVisibilityCollector):
    def factory(self, *args, **kwargs):
        return preprocess.VisibilityCollectorMem(*args, **kwargs)


class TestVisibilityCollectorHDF5(BaseTestVisibilityCollector):
    def setup(self):
        super().setup()
        self._tmpfiles = []

    def factory(self, *args, **kwargs):
        handle, filename = tempfile.mkstemp(suffix='.h5')
        self._tmpfiles.append(filename)
        os.close(handle)
        return preprocess.VisibilityCollectorHDF5(filename, *args, **kwargs)

    def teardown(self):
        for filename in self._tmpfiles:
            try:
                os.remove(filename)
            except OSError as e:
                logging.warning("Failed to remove {}: {}".format(filename, e))
        self._tmpfiles = []
